# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
Created on Sun Feb 10 13:26:58 2019

@author: Nick Anthony
"""
from __future__ import annotations

import logging
import os
import psutil
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QApplication, QMessageBox, QSplashScreen
from pwspy_gui import __version__ as version
from pwspy_gui.PWSAnalysisApp._roiManager import _DefaultROIManager, ROIManager
from pwspy_gui.PWSAnalysisApp.utilities import BlinderDialog, RoiConverter
from .dialogs import AnalysisSummaryDisplay
from ._taskManagers.analysisManager import AnalysisManager
from .mainWindow import PWSWindow
from . import applicationVars
from . import resources
from pwspy_gui.sharedWidgets.extraReflectionManager import ERManager
from typing import List
import typing
import pwspy.dataTypes as pwsdt


class PWSApp(QApplication):
    def __init__(self, args):
        logger = logging.getLogger(__name__)
        logger.debug("About to call PWSApp superclass constructor")
        super().__init__(args)
        logger.debug("PWSApp superclass constructor is finished")

        self.roiManager: ROIManager = _DefaultROIManager()

        self.setApplicationName(f"PWS Analysis v{version.split('-')[0]}")
        splash = QSplashScreen(QPixmap(os.path.join(resources, 'pwsLogo.png')))
        splash.show()
        logger.debug("Initialize ERManager")
        self.ERManager = ERManager(applicationVars.extraReflectionDirectory)
        logger.debug("Finish constructing ERManager")
        self.window = PWSWindow(self.ERManager, self.roiManager)
        splash.finish(self.window)
        logger.debug("Finish constructing window")
        self.anMan = AnalysisManager(self)
        self.window.runAction.connect(self.anMan.runList)
        availableRamGigs = psutil.virtual_memory().available / 1024**3
        if availableRamGigs > 16:  # Default to parallel analysis if we have more than 16 Gb of ram available.
            self.parallelProcessing = True  # Determines if analysis and compilation should be run in parallel or not.
        else:
            self.parallelProcessing = False  # Determines if analysis and compilation should be run in parallel or not.
        self.window.parallelAction.setChecked(self.parallelProcessing)
        self.window.parallelAction.toggled.connect(lambda checked: setattr(self, 'parallelProcessing', checked))
        logger.info(f"Initializing with useParallel set to {self.parallelProcessing}.")
        self.anMan.analysisDone.connect(lambda name, settings, warningList: AnalysisSummaryDisplay(self.window, warningList, name, settings))
        self.window.fileDialog.directoryChanged.connect(self.changeDirectory)
        self.window.blindAction.triggered.connect(self.openBlindingDialog)
        self.window.roiConvertAction.triggered.connect(self.convertRois)
        self.workingDirectory = None

        # import qdarkstyle  # This looks bad
        # dark_stylesheet = qdarkstyle.load_stylesheet_pyqt5()
        # self.setStyleSheet(dark_stylesheet)

        self.window.show()

    ### API
    def changeDirectory(self, directory: str, recursive: bool):
        from glob import glob
        pattern = [os.path.join('**', 'Cell[0-9]*')] if recursive else ['Cell[0-9]*']
        files = []
        for patt in pattern:
            files.extend(glob(os.path.join(directory, patt), recursive=recursive))
        if len(files) == 0:
            QMessageBox.information(self.window, "Hmm", "No PWS files were found.")
            return
        nums = []
        newFiles = []
        for f in files:
            try:
                nums.append(int(os.path.split(f)[-1].split('Cell')[-1]))
                newFiles.append(f)
            except ValueError:
                pass
        nums, files = zip(*sorted(zip(nums, newFiles)))
        files = list(files)
        # Load Cells
        self.window.cellSelector.loadNewCells(files, directory)
        self.workingDirectory = directory
        #Change title
        self.window.setWindowTitle(f'{QApplication.instance().applicationName()} - {directory}')
        self.workingDirectory = directory

    def openBlindingDialog(self):
        metas = self.window.cellSelector.getSelectedCellMetas()
        if len(metas) == 0:
            QMessageBox.information(self.window, "No Cells Selected", "Please select cells to act upon.")
            return
        dialog = BlinderDialog(self.window, self.workingDirectory, metas)
        dialog.exec()

    def convertRois(self):
        metas = self.window.cellSelector.getSelectedCellMetas()
        if len(metas) == 0:
            QMessageBox.information(self.window, "No Cells Selected", "Please select cells to act upon.")
            return
        rc = RoiConverter(metas)
        self.window.cellSelector.refreshCellItems()

    def setSelectedCells(self, acqs: typing.Sequence[pwsdt.Acquisition]):
        self.window.cellSelector.setSelectedCells(acqs)

    def getSelectedCells(self) -> typing.Sequence[pwsdt.Acquisition]:
        return self.window.cellSelector.getSelectedCellMetas()

    def getLoadedCells(self) -> typing.Sequence[pwsdt.Acquisition]:
        return self.window.cellSelector.getAllCellMetas()

    def plotSelectedCells(self, analysisName: str = None):
        if analysisName is None:
            analysisName = ''
        self.window.plots.setAnalysisName(analysisName)
        self.window.plots.refreshPlots()

    def __del__(self):
        try:
            self.roiManager.close()
        except AttributeError:  # If an error occured in the constructor the attribute may have never been created.
            pass



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
import os
import typing
from io import IOBase
from typing import Optional, Dict, List
import pickle
from PyQt5 import QtCore
from PyQt5.QtCore import QObject, QThread
from PyQt5.QtWidgets import QMessageBox, QWidget
import httplib2
from googleapiclient.discovery import build
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from googleapiclient.http import MediaIoBaseDownload, MediaFileUpload
from google.auth.exceptions import TransportError
from pwspy.dataTypes import ERMetaData
from pwspy_gui.PWSAnalysisApp import applicationVars
from pwspy_gui.sharedWidgets.dialogs import BusyDialog
from pwspy_gui.sharedWidgets.extraReflectionManager.ERDataComparator import ERDataComparator
from pwspy_gui.sharedWidgets.extraReflectionManager._ERDataDirectory import ERDataDirectory, EROnlineDirectory
from ._ERSelectorWindow import ERSelectorWindow
from ._ERUploaderWindow import ERUploaderWindow
from .exceptions import OfflineError

if typing.TYPE_CHECKING:
    from google.oauth2.credentials import Credentials


def _offlineDecorator(func):
    """Functions decorated with this will raise an OfflineError if they are attempted to be called while the ERManager
    is in offline mode. Only works on instance methods."""
    def wrappedFunc(self, *args, **kwargs):
        if self.offlineMode:
            logging.getLogger(__name__).warning("Attempting to download when ERManager is in offline mode.")
            raise OfflineError("Is Offline")
        func(self, *args, **kwargs)
    return wrappedFunc


class ERManager:
    """This class expects that the google drive application will already have access to a folder named
    `PWSAnalysisAppHostedFiles` which contains a folder `ExtraReflectanceCubes`, you will
    have to create these manually if starting on a new Drive account.

    Args:
        filePath: The file path to the local folder where Extra Reflection calibration files are stored.
        parentWidget: An optional reference to a QT widget that will act as the parent to any dialog windows that are opened.
    """
    def __init__(self, filePath: str, parentWidget: QWidget = None):
        self._directory = filePath
        self.offlineMode, self._downloader = self._logIn(parentWidget)

        indexPath = os.path.join(self._directory, 'index.json')
        if not os.path.exists(indexPath) and not self.offlineMode:
            self.download('index.json')
        self.localDirectory = ERDataDirectory(self._directory)
        if not self.offlineMode:
            self.onlineDirectory = EROnlineDirectory(self._downloader)
        else:
            self.onlineDirectory = None

    def _logIn(self, parentWidget: Optional[QWidget]) -> typing.Tuple[bool, Optional[ERDownloader]]:
        """
        Try logging in to google drive
        Args:
            parentWidget: If a message box needs to be displayed this widget will act as the parent of the message box.

        Returns:
            offlineMode: True if connection failed.
            downloader: A download helper object. Will be none if connection failed.

        """
        logger = logging.getLogger(__name__)
        logger.debug("Calling ERDownloader.getCredentials")
        creds = ERDownloader.getCredentials(applicationVars.googleDriveAuthPath)
        logger.debug("Finished ERDownloader.getCredentials")
        if creds is None:  # Check if the google drive credentials exists and if they don't then give the user a message.
            msg = QMessageBox(parentWidget)
            msg.setIcon(QMessageBox.Information)
            msg.setText("Please log in to the google drive account containing the PWS Calibration Database. This is currently backman.lab@gmail.com")
            msg.setWindowTitle("Time to log in!")
            msg.setWindowModality(QtCore.Qt.WindowModal)
            okButton = msg.addButton("Ok", QMessageBox.YesRole)
            skipButton = msg.addButton("Skip (offline mode)", QMessageBox.NoRole)
            msg.exec()
            if msg.clickedButton() is skipButton:
                return True, None
        try:
            downloader = ERDownloader(applicationVars.googleDriveAuthPath)
            return False, downloader
        except (TransportError, httplib2.ServerNotFoundError):
            msg = QMessageBox.information(parentWidget, "Internet?", "Google Drive connection failed. Proceeding in offline mode.")
            return True, None

    def createSelectorWindow(self, parent: QWidget):
        return ERSelectorWindow(self, parent)

    def createManagerWindow(self, parent: QWidget):
        return ERUploaderWindow(self, parent)

    @_offlineDecorator
    def download(self, fileName: str, parentWidget: Optional[QWidget] = None):
        """Begin downloading `fileName` in a separate thread. Use the main thread to update a progress bar.
        If directory is left blank then file will be downloaded to the ERManager main directory"""
        self._downloader.download(fileName, self._directory, parentWidget)

    @_offlineDecorator
    def upload(self, fileName: str):
        """Uploads the file at `fileName` to the `ExtraReflectanceCubes` folder of the google drive account"""
        filePath = os.path.join(self._directory, fileName)
        self._downloader.upload(filePath)

    def getMetadataFromId(self, idTag: str) -> ERMetaData:
        """Given the unique idTag string for an ExtraReflectanceCube this will search the index.json and return the
        ERMetaData file. If it cannot be found then an `IndexError will be raised."""
        return self.localDirectory.getMetadataFromId(idTag)


class ERDownloader:
    """Implements downloading functionality specific to the structure that we have calibration files stored on our google drive account."""
    def __init__(self, authPath: str):
        self._downloader = _QtGoogleDriveDownloader(authPath)

    def download(self, fileName: str, directory: str, parentWidget: Optional[QWidget] = None):
        """Begin downloading `fileName` in a separate thread. Use the main thread to update a progress bar.
        If directory is left blank then file will be downloaded to the ERManager main directory"""
        t = self._DownloadThread(self._downloader, fileName, directory)
        b = BusyDialog(parentWidget, f"Downloading {fileName}. Please Wait...", progressBar=True)  # This dialog blocks the screen until the download thread is completed.
        t.finished.connect(b.accept)  # When the thread finishes, close the busy dialog.
        self._downloader.progress.connect(b.setProgress)  # Progress from the downloader updates a progress bar on the busy dialog.
        t.errorOccurred.connect(lambda e: QMessageBox.information(parentWidget, 'Error in Drive Downloader Thread', str(e)))
        t.start()
        b.exec()

    def downloadToRam(self, fileName: str, stream: IOBase) -> IOBase:
        """Download a file directly to a stream in ram rather than saving to file, best for small temporary files.
        Args:
            fileName (str): The name of the file stored on google drive, must be in the Extra reflectance directory.
            stream (IOBase): An empty stream that the file contents will be loaded into.
        Returns:
            IOBase: The same stream that was passed in as `stream`."""
        files = self._downloader.getFolderIdContents(
            self._downloader.getIdByName('PWSAnalysisAppHostedFiles'))
        files = self._downloader.getFolderIdContents(
            self._downloader.getIdByName('ExtraReflectanceCubes', fileList=files))
        fileId = self._downloader.getIdByName(fileName, fileList=files)  
        self._downloader.downloadFile(fileId, stream)
        return stream

    def upload(self, filePath: str):
        parentId = self._downloader.getIdByName("ExtraReflectanceCubes")
        self._downloader.uploadFile(filePath, parentId)

    def getFileMetadata(self) -> List[Dict]:
        """Return GoogleDrive metadata about the files in the extra reflectance drive folder"""
        files = self._downloader.getFolderIdContents(
            self._downloader.getIdByName('PWSAnalysisAppHostedFiles'))
        files = self._downloader.getFolderIdContents(
            self._downloader.getIdByName('ExtraReflectanceCubes', fileList=files))
        return files

    @staticmethod
    def getCredentials(authPath: str):
        return GoogleDriveDownloader.getCredentials(authPath)

    class _DownloadThread(QThread):
        """A QThread to download from google drive"""
        errorOccurred = QtCore.pyqtSignal(Exception)  # If an exception occurs it can be passed to another thread with this signal

        def __init__(self, downloader: GoogleDriveDownloader, fileName: str, directory: str):
            super().__init__()
            self.downloader = downloader
            self.fileName = fileName
            self.directory = directory

        def run(self):
            try:
                files = self.downloader.getFolderIdContents(
                    self.downloader.getIdByName('PWSAnalysisAppHostedFiles'))
                files = self.downloader.getFolderIdContents(
                    self.downloader.getIdByName('ExtraReflectanceCubes', fileList=files))
                fileId = self.downloader.getIdByName(self.fileName, fileList=files)
                with open(os.path.join(self.directory, self.fileName), 'wb') as f:
                    self.downloader.downloadFile(fileId, f)
            except Exception as e:
                self.errorOccurred.emit(e)


class GoogleDriveDownloader:
    """Handles downloading and uploading files from Google Drive.

    Upon initializing an instance of this class you will be asked for username and password if you don't already have
    authentication saved from a previous login.

    Args:
        authPath: The folder to store authentication files. Before this class will work you will need to place
            `credentials.json` in the authPath. You can get this file from the online Google Drive api console. Create
            an Oauth 2.0 credential with access to the `drive.file` api.

."""
    def __init__(self, authPath: str):
        self._allFiles = None  #A list of all the files that we have access to. updated by self._updateFilesList()
        self._authPath = authPath
        tokenPath = os.path.join(self._authPath, 'driveToken.pickle')
        credPath = os.path.join(self._authPath, 'credentials.json')
        creds = self.getCredentials(self._authPath)
        if not creds or not creds.valid:  # If there are no (valid) credentials available, let the user log in.
            if creds and creds.expired and creds.refresh_token:  # Attempt to refresh the credentials.
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(
                    credPath, ['https://www.googleapis.com/auth/drive.file'])
                creds = flow.run_local_server(port=8090)  # Opens the google login window in a browser.
            with open(tokenPath, 'wb') as token:  # Save the credentials for the next run
                pickle.dump(creds, token)
        self.api = build('drive', 'v3', credentials=creds)  # this returns access to the drive api. see google documentation. All drive related functionality happens through this api.
        self._updateFilesList()

    @staticmethod
    def getCredentials(authPath: str) -> Credentials:
        """
        Args:
            authPath: The folder path to the authentication folder.

        Returns:
             The Google Drive credentials stored in `driveToken.pickle`
        """
        tokenPath = os.path.join(authPath, 'driveToken.pickle')
        creds = None
        if os.path.exists(tokenPath):
            with open(tokenPath, 'rb') as token:
                creds = pickle.load(token)
        return creds

    def _updateFilesList(self):
        """Update the list of all files in the Google Drive account. This is automatically called during initialization
         and after uploading a new file. I don't think it should be needed anywhere else."""
        results = self.api.files().list(fields="nextPageToken, files(id, name, parents, md5Checksum)").execute()
        self._allFiles = results.get('files', [])

    def getIdByName(self, name: str, fileList: Optional[List] = None) -> str:
        """Return the file id associated with a filename.

        Args:
            name: The filename you want the ID for.
            fileList: A list of metadata such as is returned by getFolderIDContents. if left as `None` then all files
            are searched. If there are multiple files with the same name the first match that is found will be returned.

        Returns:
            The ID string of the file.
        """
        if fileList is None:
            fileList = self._allFiles
        matches = [i['id'] for i in fileList if i['name'] == name]
        if len(matches) > 1:
            raise ValueError(f"Google Drive found multiple files matching file name: {name}")
        elif len(matches) == 0:
            raise ValueError(f"Google Drive found not files matching file name: {name}")
        else:
            return matches[0]

    def getFolderIdContents(self, Id: str) -> List[Dict]:
        """Return the API metadata for all files contained within the folder associated with `id`.

        Args:
            Id: the file ID as returned by `getIdByName`.

        Returns:
            A list of file metadata.
        """
        files = [i for i in self._allFiles if 'parents' in i] # Get rid of parentless files. they will cause errors.
        return [i for i in files if Id in i['parents']]

    def downloadFile(self, Id: str, file: IOBase):
        """Save the file with `id` as it Google Drive ID

        Args:
            Id: the file ID as returned by `getIdByName`.
            file: A file or other stream to save to.
        """
        fileRequest = self.api.files().get_media(fileId=Id)
        downloader = MediaIoBaseDownload(file, fileRequest)
        done = False
        logger = logging.getLogger(__name__)
        while done is False:
            status, done = downloader.next_chunk()
            logger.info("Download %d%%." % int(status.progress() * 100))

    def createFolder(self, name: str, parentId: Optional[str] = None) -> str:
        """Creates a folder with name `name` and returns the id number of the folder
        If parentId is provided then the folder will be created inside the parent folder.

        Args:
            name: The name of the new folder
            parentId: The ID of the folder you want to create this folder inside of.

        Returns:
            The ID of the new folder
        """
        folderMetadata = {'name': name, 'mimeType': 'application/vnd.google-apps.folder'}
        if parentId: folderMetadata['parents'] = [parentId]
        folder = self.api.files().create(body=folderMetadata, fields='id').execute()
        return folder.get('id')

    def uploadFile(self, filePath: str, parentId: str) -> str:
        """Upload the file at `filePath` to the folder with `parentId`, keeping the original file name.
        If a file with the same parent and name already exists, replace it.

        Args:
            filePath: the local path the file that should be uploaded
            parentId: The ID of the folder that the file should be uploaded to.

        Returns:
            The ID of the newly uploaded file.

        """
        fileName = os.path.split(filePath)[-1]
        existingFiles = self.getFolderIdContents(parentId)
        if fileName in [i['name'] for i in existingFiles]: #FileName already exists
            existingId = [i['id'] for i in existingFiles if i['name'] == fileName][0]
            self.api.files().delete(fileId=existingId).execute()
        fileMetadata = {'name': fileName, 'parents': [parentId]}
        media = MediaFileUpload(filePath)
        file = self.api.files().create(body=fileMetadata, media_body=media, fields='id').execute()
        self._updateFilesList()
        return file.get('id')

    def moveFile(self, fileId: str, newFolderId: str):
        """Move a file that is already uploaded to Google Drive.

        Args:
            fileId: the ID of the file that should be moved.
            newFolderId: The ID of the parent folder you want to move the file to.

        """
        file = self.api.files().get(fileId=fileId, fields='parents').execute()
        oldParents = ','.join(file.get('parents'))
        file = self.api.files().update(fileId=fileId, addParents=newFolderId, removeParents=oldParents, fields='id, parents').execute()


class _QtGoogleDriveDownloader(GoogleDriveDownloader, QObject):
    """Same as the standard google drive downloader except it emits a progress signal after each chunk downloaded. This can be used to update a progress bar."""
    progress = QtCore.pyqtSignal(int)  # gives an estimate of download progress percentage

    def __init__(self, authPath: str):
        GoogleDriveDownloader.__init__(self, authPath)
        QObject.__init__(self)

    def downloadFile(self, Id: str, file: IOBase):
        """Save the file with googledrive file identifier `Id` to `savePath` while emitting the `progress` signal
        which can be connected to a progress bar or whatever."""
        fileRequest = self.api.files().get_media(fileId=Id)
        downloader = MediaIoBaseDownload(file, fileRequest, chunksize=1024*1024*5) # Downloading in 5mb chunks instead of the default 100mb chunk just so that the progress bar looks smoother
        done = False
        while done is False:
            status, done = downloader.next_chunk()
            self.progress.emit(int(status.progress() * 100))



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
import os
from glob import glob
import typing

from pwspy_gui.PWSAnalysisApp._dockWidgets.AnalysisSettingsDock.widgets.SettingsFrames._AbstractSettingsFrame import AbstractSettingsFrame
from pwspy_gui.PWSAnalysisApp._dockWidgets.AnalysisSettingsDock.widgets.SettingsFrames._sharedWidgets import ExtraReflectanceSelector, HardwareCorrections, \
    QHSpinBox, QHDoubleSpinBox, VerticallyCompressedWidget
from pwspy_gui.PWSAnalysisApp.componentInterfaces import CellSelector
from pwspy_gui.PWSAnalysisApp._dockWidgets.AnalysisSettingsDock.runtimeSettings import PWSRuntimeAnalysisSettings
if typing.TYPE_CHECKING:
    from pwspy_gui.sharedWidgets.extraReflectionManager import ERManager

from PyQt5 import QtCore, QtGui
from PyQt5.QtWidgets import QScrollArea, QGridLayout, QLineEdit, QLabel, QGroupBox, QHBoxLayout, QWidget, QRadioButton, \
    QFrame, QCheckBox, QDoubleSpinBox

from pwspy.analysis.pws import PWSAnalysisSettings
from pwspy_gui.PWSAnalysisApp import applicationVars
from pwspy_gui.PWSAnalysisApp.sharedWidgets.collapsibleSection import CollapsibleSection


class PWSSettingsFrame(AbstractSettingsFrame, QScrollArea):
    def __init__(self, erManager: ERManager, cellSelector: CellSelector):
        super().__init__()
        self.cellSelector = cellSelector

        self._frame = VerticallyCompressedWidget(self)
        self._layout = QGridLayout()
        self._frame.setLayout(self._layout)
        self._frame.setFixedWidth(350)
        self.setMinimumWidth(self._frame.width()+5)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setWidget(self._frame)

        """Set up Frame"""
        """Presets"""
        row = 0
        self._analysisNameEdit = QLineEdit()
        self._layout.addWidget(QLabel("Analysis Name: "), row, 0, 1, 1)
        self._layout.addWidget(self._analysisNameEdit, row, 1, 1, 1)
        row += 1
        self.presets = QGroupBox("Presets")
        self.presets.setLayout(QHBoxLayout())
        self.presets.layout().setContentsMargins(0, 0, 0, 5)
        _2 = QWidget()
        _2.setLayout(QHBoxLayout())
        _2.layout().setContentsMargins(5, 0, 5, 0)
        for f in glob(os.path.join(applicationVars.analysisSettingsDirectory, '*_analysis.json')):
            name = os.path.split(f)[-1][:-14]
            b = QRadioButton(name)
            b.released.connect(
                lambda n=name: self.loadFromSettings(
                    PWSAnalysisSettings.fromJson(applicationVars.analysisSettingsDirectory, n)))
            _2.layout().addWidget(b)
        _ = QScrollArea()
        _.setWidget(_2)
        _.setFrameShape(QFrame.NoFrame)
        _.setContentsMargins(0, 0, 0, 0)
        _.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        _.horizontalScrollBar().setStyleSheet("QScrollBar:horizontal { height: 10px; }")
        self.presets.setFixedHeight(45)
        self.presets.layout().addWidget(_)
        self._layout.addWidget(self.presets, row, 0, 1, 4)
        row += 1

        '''Hardwarecorrections'''
        self.hardwareCorrections = HardwareCorrections(self)
        self.hardwareCorrections.stateChanged.connect(self._updateSize)
        self._layout.addWidget(self.hardwareCorrections, row, 0, 1, 4)
        row += 1

        '''Extra Reflection'''
        self.extraReflection = ExtraReflectanceSelector(self, erManager)
        self._layout.addWidget(self.extraReflection, row, 0, 1, 4)
        row += 1

        '''Use relative or absolute units of reflectance'''
        self.scaling = QGroupBox("Scaling")
        self.scaling.setLayout(QHBoxLayout())
        self.scaling.layout().setContentsMargins(2, 2, 2, 5)
        self.relativeUnits = QCheckBox("Use Relative Units", self)
        self.relativeUnits.setToolTip("If checked then reflectance (and therefore all other parameters) will be scaled such that any reflectance matching that\n"
                                      "of the reference image will be 1. If left unchecked then the `Reference Material` will be used to scale reflectance to\n"
                                      "match the actual physical reflectance of the sample.")
        self.scaling.layout().addWidget(self.relativeUnits)
        self._layout.addWidget(self.scaling, row, 0, 1, 4)
        row += 1

        '''SignalPreparations'''
        self.signalPrep = QGroupBox("Signal Prep")
        self.signalPrep.setFixedSize(175, 75)
        self.signalPrep.setToolTip("In order to reduce the effects of measurement noise we filter out the high frequencies from our signal. We do this using a\n"
                                   "Buttersworth low-pass filter. Best to stick with the defaults on this one.")
        layout = QGridLayout()
        layout.setContentsMargins(5, 1, 5, 5)
        _ = layout.addWidget

        orderLabel = QLabel("Filter Order")
        self.filterOrder = QHSpinBox()
        self.filterOrder.setRange(0, 6)
        self.filterOrder.setToolTip("A low-pass filter is applied to the spectral signal to reduce noise. This determines the `order` of the digital filter.")
        orderLabel.setToolTip(self.filterOrder.toolTip())
        cutoffLabel = QLabel("Cutoff Freq.")
        self.filterCutoff: QDoubleSpinBox = QHDoubleSpinBox()
        self.filterCutoff.setToolTip("The frequency in units of 1/wavelength for the filter cutoff.")
        self.filterCutoff.setDecimals(3)  # Greater precision than default
        cutoffLabel.setToolTip(self.filterCutoff.toolTip())
        self.filterCheckbox = QCheckBox("Low-pass Filtering", self)
        def stateChange(state: bool):
            self.filterOrder.setEnabled(state)
            self.filterCutoff.setEnabled(state)
        self.filterCheckbox.stateChanged.connect(stateChange)
        self.filterCheckbox.setChecked(True)  # This is just to initialize the proper state
        _(self.filterCheckbox, 0, 0, 1, 2)
        _(orderLabel, 1, 0, 1, 1)
        _(self.filterOrder, 1, 1, 1, 1)
        _(cutoffLabel, 2, 0, 1, 1)
        _(self.filterCutoff, 2, 1, 1, 1)
        _(QLabel("nm<sup>-1</sup>"), 2, 2, 1, 1)
        self.signalPrep.setLayout(layout)
        self._layout.addWidget(self.signalPrep, row, 0, 1, 2)

        '''Cropping'''
        self.cropping = QGroupBox("Wavelength Cropping")
        self.cropping.setFixedSize(125, 75)
        self.cropping.setToolTip("In the past it was found that there was exceptionally high noise at the very beginning and end of an acquisition. For this reason we would exclude the first and last wavelengths of the image cube. While it is likely that the noise issue has now been fixed we still do this for consistency's sake.")
        layout = QGridLayout()
        layout.setContentsMargins(5, 1, 5, 5)
        _ = layout.addWidget
        self.wavelengthStart = QHSpinBox()
        self.wavelengthStop = QHSpinBox()
        self.wavelengthStart.setToolTip("Sometimes the beginning and end of the spectrum can have very high noise. For this reason we crop the data before analysis.")
        self.wavelengthStop.setToolTip("Sometimes the beginning and end of the spectrum can have very high noise. For this reason we crop the data before analysis.")
        self.wavelengthStart.setRange(300, 800)
        self.wavelengthStop.setRange(300, 800)
        self.croppingCheckbox = QCheckBox("Enable Cropping", self)
        def cropStateChanged(state: bool):
            self.wavelengthStop.setEnabled(state)
            self.wavelengthStart.setEnabled(state)
        self.croppingCheckbox.stateChanged.connect(cropStateChanged)
        self.croppingCheckbox.setChecked(True)  # This is just to initialze the propert state.
        _(self.croppingCheckbox, 0, 0, 1, 2)
        _(QLabel("Start"), 1, 0)
        _(QLabel("Stop"), 1, 1)
        _(self.wavelengthStart, 2, 0)
        _(self.wavelengthStop, 2, 1)
        self.cropping.setLayout(layout)
        self._layout.addWidget(self.cropping, row, 2, 1, 2)
        row += 1

        '''Polynomial subtraction'''
        self.polySub = QGroupBox("Polynomial Subtraction")
        self.polySub.setFixedSize(150, 50)
        self.polySub.setToolTip("A polynomial is fit to each spectrum and then it is subtracted from the spectrum."
                                "This is so that we remove effects of absorbtion and our final signal is only due to interference"
                                "For liquid covered samples this is always set to 0. For samples in air a 2nd order polynomial is used.")
        layout = QGridLayout()
        layout.setContentsMargins(5, 1, 5, 5)
        _ = layout.addWidget
        self.polynomialOrder = QHSpinBox()
        _(QLabel("Order"), 0, 0, 1, 1)
        _(self.polynomialOrder, 0, 1, 1, 1)
        self.polySub.setLayout(layout)
        self._layout.addWidget(self.polySub, row, 0, 1, 2)

        # WaveNumber Filtering
        self.waveNumberCutoff = QHDoubleSpinBox()
        self.waveNumberCutoff.setValue(4.3)  # OPD = estimated depth * 2(roundtrip) * 1.37 (RI of cell). DOF=1.1um -> OPD = 2.95. To achieve 5% attenuation at 2.95um a 3dB cutoff of 4.3 is selected (4th-order butterworth)
        self.cutoffWaveNumber = QCheckBox("Perform Filtering")
        self.cutoffWaveNumber.stateChanged.connect(lambda state: self.waveNumberCutoff.setEnabled(state != QtCore.Qt.Unchecked))
        self.cutoffWaveNumber.setCheckState(QtCore.Qt.Checked)
        self.cutoffWaveNumber.setCheckState(QtCore.Qt.Unchecked)  # Default to off

        layout = QGridLayout()
        layout.setContentsMargins(5, 1, 5, 5)

        layout.addWidget(self.cutoffWaveNumber, 0, 0, 1, 2)
        layout.addWidget(QLabel("OPD Cutoff (um)"), 1, 0, 1, 1)
        layout.addWidget(self.waveNumberCutoff, 1, 1, 1, 1)
        wnFilter = QGroupBox("OPD Filtering")
        wnFilter.setLayout(layout)
        wnFilter.setToolTip("A 4th order buttersworth filter will be applied to the spectra (x axis: wavenumber) at the specified cutoff frequency."
                            "The cutoff is the frequency at which the attenuation will be -3dB. After the cutoff the attenuation slope will be 80dB per decade.")
        self._layout.addWidget(wnFilter, row, 2, 1, 2)

        row += 1


        '''Advanced Calculations'''
        self.advanced = CollapsibleSection('Skip Advanced Analysis', 200, self)
        self.advanced.stateChanged.connect(self._updateSize)
        self.advanced.setToolTip("If this box is ticked then some of the less common analyses will be skipped. This saves time and harddrive space.")
        self.autoCorrStopIndex = QHSpinBox()
        self.autoCorrStopIndex.setToolTip("Autocorrelation slope is determined by fitting a line to the first values of the autocorrelation function. This value determines how many values to include in this linear fit.")
        self.minSubCheckBox = QCheckBox("MinSub")
        self.minSubCheckBox.setToolTip("The calculation of autocorrelation decay slope involves taking the natural logarithm of of the autocorrelation. However noise often causes the autocorrelation to have negative values which causes problems for the logarithm. Checking this box adds an offset to the autocorrelation so that no values are negative.")
        layout = QGridLayout()
        _ = layout.addWidget
        _(QLabel("AutoCorr Stop Index"), 0, 0, 1, 1)
        _(self.autoCorrStopIndex, 0, 1, 1, 1)
        _(self.minSubCheckBox, 1, 0, 1, 1)
        self.advanced.setLayout(layout)
        self._layout.addWidget(self.advanced, row, 0, 1, 4)
        row += 1

        self._updateSize()

    def showEvent(self, a0: QtGui.QShowEvent) -> None:
        super().showEvent(a0)
        self._updateSize() #For some reason this must be done here and in the __init__ for it to start up properly.

    def _updateSize(self):
        height = 100  # give this much excess room.
        height += self.presets.height()
        height += self.hardwareCorrections.height()
        height += self.extraReflection.height()
        height += self.scaling.height()
        height += self.signalPrep.height()
        height += self.polySub.height()
        height += self.advanced.height()
        self._frame.setFixedHeight(height)

    def loadFromSettings(self, settings: PWSAnalysisSettings):
        if settings.filterCutoff is None:
            self.filterCheckbox.setChecked(False)
        else:
            self.filterCheckbox.setChecked(True)
            self.filterCutoff.setValue(settings.filterCutoff)
        self.filterOrder.setValue(settings.filterOrder)
        self.polynomialOrder.setValue(settings.polynomialOrder)
        self.extraReflection.loadFromSettings(settings.numericalAperture, settings.referenceMaterial, settings.extraReflectanceId)
        if settings.wavelengthStop is None:
            self.croppingCheckbox.setChecked(False)
        else:
            self.croppingCheckbox.setChecked(True)
            self.wavelengthStop.setValue(settings.wavelengthStop)
            self.wavelengthStart.setValue(settings.wavelengthStart)
        self.advanced.setCheckState(2 if settings.skipAdvanced else 0)
        self.autoCorrStopIndex.setValue(settings.autoCorrStopIndex)
        self.minSubCheckBox.setCheckState(2 if settings.autoCorrMinSub else 0)
        self.relativeUnits.setCheckState(2 if settings.relativeUnits else 0)
        self.hardwareCorrections.loadCameraCorrection(settings.cameraCorrection)
        if settings.waveNumberCutoff is None:
            self.cutoffWaveNumber.setCheckState(QtCore.Qt.Unchecked)
        else:
            self.cutoffWaveNumber.setCheckState(QtCore.Qt.Checked)
            self.waveNumberCutoff.setValue(settings.waveNumberCutoff)

    def getSettings(self) -> PWSRuntimeAnalysisSettings:
        erMetadata, refMaterial, numericalAperture = self.extraReflection.getSettings()
        refMeta = self.cellSelector.getSelectedReferenceMeta()
        cellMeta = self.cellSelector.getSelectedCellMetas()
        name = self._analysisNameEdit.text()
        if refMeta is None:
            raise ValueError('Please select a reference Cell.')
        if name == '':
            raise ValueError("Please give your analysis a name.")
        if len(cellMeta) == 0:
            raise ValueError('Please select cells to analyse.')
        refMeta = refMeta.pws
        cellMeta = [i.pws for i in cellMeta if i.pws is not None]  # If we select some acquisitions that don't have dynamics then they'll get stripped out here
        if refMeta is None:
            raise ValueError("The selected reference acquisition has no valid PWS data.")
        if len(cellMeta) == 0:
            raise ValueError("No valid PWS acquisitions were selected.")
        cutoff = self.filterCutoff.value() if self.filterCheckbox.checkState() else None
        wvStart = self.wavelengthStart.value() if self.croppingCheckbox.checkState() else None
        wvStop = self.wavelengthStop.value() if self.croppingCheckbox.checkState() else None
        return PWSRuntimeAnalysisSettings(settings=PWSAnalysisSettings(filterOrder=self.filterOrder.value(),
                                                                       filterCutoff=cutoff,
                                                                       polynomialOrder=self.polynomialOrder.value(),
                                                                       referenceMaterial=refMaterial,
                                                                       wavelengthStart=wvStart,
                                                                       wavelengthStop=wvStop,
                                                                       skipAdvanced=self.advanced.checkState() != 0,
                                                                       autoCorrMinSub=self.minSubCheckBox.checkState() != 0,
                                                                       autoCorrStopIndex=self.autoCorrStopIndex.value(),
                                                                       numericalAperture=numericalAperture,
                                                                       relativeUnits=self.relativeUnits.checkState() != 0,
                                                                       cameraCorrection=self.hardwareCorrections.getCameraCorrection(),
                                                                       extraReflectanceId=erMetadata.idTag if erMetadata is not None else None,
                                                                       waveNumberCutoff=self.waveNumberCutoff.value() if self.cutoffWaveNumber.checkState != QtCore.Qt.Unchecked else None),
                                          extraReflectanceMetadata=erMetadata,
                                          referenceMetadata=refMeta,
                                          cellMetadata=cellMeta,
                                          analysisName=name)






# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
import os
import typing as t_
from PyQt5 import QtGui, QtCore
from PyQt5.QtGui import QPalette, QValidator, QDoubleValidator
from PyQt5.QtWidgets import QGroupBox, QWidget, QVBoxLayout, QLabel, QLineEdit, QPushButton, QHBoxLayout, QMessageBox, QGridLayout, QSpinBox, QDoubleSpinBox, \
    QComboBox, QFrame, QSpacerItem, QSizePolicy, QLayout

from pwspy_gui import resources
from pwspy_gui.PWSAnalysisApp.sharedWidgets import CollapsibleSection
from pwspy.dataTypes import CameraCorrection, ERMetaData
from pwspy.utility.reflection import reflectanceHelper, Material
if t_.TYPE_CHECKING:
    from pwspy_gui.sharedWidgets.extraReflectionManager import ERManager


class ExtraReflectanceSelector(QGroupBox):
    def __init__(self, parent: QWidget, erManager: ERManager):
        super().__init__("Extra Reflection", parent)
        self.ERExplorer = None # We delay the instantiation of this window because its __init__ triggers some somewhat slow processes that aren't needed if the window is never opened.
        self._erManager = erManager

        self.setToolTip("The fact that some light captured by the camera is scattered off surfaces inside the objective without ever reaching the sample means\n"
                        "that the light intensity captured by the camera is not proportional to the reflectance of the sample. This extra reflectance varies\n"
                        "spatially and spectrally and must be subtracted from our data in order for the analysis to be accurate. Calibration measurements of\n"
                        "the extra reflectance are periodically uploaded to a google drive account which this software can download from. Click the folder icon\n"
                        "to view and download the available calibration data cubes. Make sure to select one for the correct system and a date that is close to\n"
                        "the acquisition date of your data. The `reference material` should be selected to match the material that was imaged in your reference\n"
                        "image cube (usually Water).")
        layout = QVBoxLayout()
        layout.setContentsMargins(5, 1, 5, 5)
        rsubLabel = QLabel("R Subtraction")
        self.RSubtractionNameLabel = QLineEdit('None')
        self.RSubtractionNameLabel.setReadOnly(True)
        self.RSubtractionBrowseButton = QPushButton(QtGui.QIcon(os.path.join(resources, 'folder.svg')), '')
        self.RSubtractionBrowseButton.released.connect(self._browseReflection)
        refMatLabel = QLabel("Reference Material")
        self.refMaterialCombo = QHComboBox()
        self.refMaterialCombo.addItems([k.name for k in reflectanceHelper.materialFiles.keys() if k.name != 'Glass'])
        naLabel = QLabel("Numerical Aperture")
        self.numericalAperture = QHDoubleSpinBox()
        self.numericalAperture.setRange(0, 2)
        self.numericalAperture.setSingleStep(0.1)
        self.numericalAperture.setToolTip("The illumination numerical aperture used. This is usually 0.52 on NU systems."
                                          "This is used to accurately calculate the theoretically expected reflectance of "
                                          "the reference material. We also want to check that the ExtraReflection was taken "
                                          "at the same NA.")
        naLabel.setToolTip(self.numericalAperture.toolTip())
        rLayout = QHBoxLayout()
        _ = rLayout.addWidget
        _(rsubLabel)
        _(self.RSubtractionNameLabel)
        _(self.RSubtractionBrowseButton)
        layout.addLayout(rLayout)
        rLayout = QHBoxLayout()
        _ = rLayout.addWidget
        _(refMatLabel)
        _(self.refMaterialCombo)
        layout.addLayout(rLayout)
        rLayout = QHBoxLayout()
        _ = rLayout.addWidget
        _(naLabel)
        _(self.numericalAperture)
        layout.addLayout(rLayout)
        self.setLayout(layout)

    def getSettings(self) -> t_.Tuple[ERMetaData, Material, float]:
        self._initializeERSelector()
        if self.ERExplorer.getSelectedMetadata() is None:
            ans = QMessageBox.question(self, "Uh", "An extra reflectance cube has not been selected. Do you want to ignore this important correction?")
            if ans == QMessageBox.Yes:
                erMd = None
            else:
                raise ValueError("An extra reflectance cube has not been selected.")
        else:
            erMd = self.ERExplorer.getSelectedMetadata()
        self.numericalAperture.clearFocus() #This should prevent an error where the value that is saved doesn't match what is actually typed in when the keyboard is still focused on the spinbox.
        numericalAperture = self.numericalAperture.value()
        refMaterial = Material[self.refMaterialCombo.currentText()]
        return erMd, refMaterial, numericalAperture

    def getSelectedERMetadata(self) -> t_.Optional[ERMetaData]:
        return self.ERExplorer.getSelectedMetadata()

    def loadFromSettings(self, numericalAperture: float, referenceMaterial: Material, extraReflectanceId: str):
        if extraReflectanceId is not None:
            self._initializeERSelector()
            try:
                md = self._erManager.getMetadataFromId(extraReflectanceId)
                self.ERExplorer.setSelection(md)
            except IndexError:  # Metadata matching that ID was not found by the ERManager
                self.RSubtractionNameLabel.setText(f"Failed to find ID: {extraReflectanceId}")
        if referenceMaterial is None:  # Even though choosing a referenceMaterial of None is no longer an option, it was in the past, we still support loading these settings.
            self.refMaterialCombo.setCurrentText("Ignore")
        else:
            self.refMaterialCombo.setCurrentIndex(self.refMaterialCombo.findText(referenceMaterial.name))
        self.numericalAperture.setValue(numericalAperture)

    def _browseReflection(self):
        self._initializeERSelector()
        self.ERExplorer.show()

    def _initializeERSelector(self):
        if self.ERExplorer is None:  # Don't initialize the ERSelector window until we actually need it. It can be slow to initiate
            self.ERExplorer = self._erManager.createSelectorWindow(self)

            def extraReflectionChanged(md: t_.Optional[ERMetaData]):
                if md is None:
                    self.RSubtractionNameLabel.setText('None')
                else:
                    self.RSubtractionNameLabel.setText(os.path.split(md.filePath)[-1])

            self.ERExplorer.selectionChanged.connect(extraReflectionChanged)

class HardwareCorrections(CollapsibleSection):
    def __init__(self, parent: QWidget):
        super().__init__('Automatic Correction', 200, parent)
        layout = QGridLayout()
        dcLabel = QLabel('Dark Counts')
        self.darkCountBox = QHSpinBox()
        self.darkCountBox.setToolTip("The counts/pixel reported by the camera when it is not exposed to any light."
                                     " e.g if using 2x2 binning and you measure 400 counts, then the value to put here is 100.")
        dcLabel.setToolTip(self.darkCountBox.toolTip())
        self.darkCountBox.setRange(0, 10000)
        linLabel = QLabel("Linearity Correction")
        self.linearityEdit = QLineEdit()
        self.linearityEdit.setText("1")
        self.linearityEdit.setToolTip("A comma-separated polynomial to linearize the counts from the camera."
                                      "E.G an entry of A,B,C here will result in the data being transformed as newData = A * data + B * data^2 + C * data^3."
                                      "Leaving this as '1' will result in no transformation (usually CMOS cameras are already linear)")
        linLabel.setToolTip(self.linearityEdit.toolTip())
        self.linearityEdit.setValidator(HardwareCorrections.CsvValidator())
        origPalette = self.linearityEdit.palette()
        palette = QPalette()
        palette.setColor(QPalette.Text, QtCore.Qt.red)
        self.linearityEdit.validator().stateChanged.connect(lambda state:
                                                            self.linearityEdit.setPalette(
                                                                palette) if state != QValidator.Acceptable else self.linearityEdit.setPalette(origPalette))

        _ = layout.addWidget
        _(dcLabel, 0, 0)
        _(self.darkCountBox, 0, 1)
        _(linLabel, 1, 0)
        _(self.linearityEdit, 1, 1)

        self.setToolTip("The relationship between camera counts and light intensity is not always linear."
                                            "The correction parameters can usually be found automatically in the image metadata.")
        self.setLayout(layout)

    def getCameraCorrection(self) -> CameraCorrection:
        if self.checkState() == 0:
            if self.linearityEdit.validator().state != QValidator.Acceptable:
                raise ValueError("The camera linearity correction input is not valid.")
            linText = self.linearityEdit.text()
            linearityPoly = tuple(float(i) for i in linText.split(','))
            cameraCorrection = CameraCorrection(self.darkCountBox.value(), linearityPoly)
        else:
            cameraCorrection = None
        return cameraCorrection

    def loadCameraCorrection(self, camCorr: t_.Optional[CameraCorrection] = None):
        if camCorr is None: #Automatic camera corrections
            self.setCheckState(2)
        else:
            self.setCheckState(0)
            if camCorr.linearityPolynomial is None:
                self.linearityEdit.setText("1")
            else:
                self.linearityEdit.setText(",".join((str(i) for i in camCorr.linearityPolynomial)))
            self.darkCountBox.setValue(camCorr.darkCounts)

    class CsvValidator(QValidator):
        stateChanged = QtCore.pyqtSignal(QValidator.State)

        def __init__(self):
            super().__init__()
            self.doubleValidator = QDoubleValidator()
            self.state = QValidator.Acceptable

        def validate(self, inp: str, pos: int):
            oldState = self.state
            for i in inp.split(','):
                ret = self.doubleValidator.validate(i, pos)
                if ret[0] == QValidator.Intermediate:
                    self.state = ret[0]
                    if self.state != oldState: self.stateChanged.emit(self.state)
                    return self.state, inp, pos
                elif ret[0] == QValidator.Invalid:
                    return ret
            self.state = QValidator.Acceptable
            if self.state != oldState: self.stateChanged.emit(self.state)
            return self.state, inp, pos


def humble(clas):
    """Returns a subclass of clas that will not allow scrolling unless it has been actively selected."""
    class HumbleDoubleSpinBox(clas):
        def __init__(self, *args):
            super(HumbleDoubleSpinBox, self).__init__(*args)
            self.setFocusPolicy(QtCore.Qt.StrongFocus)

        def focusInEvent(self, event):
            self.setFocusPolicy(QtCore.Qt.WheelFocus)
            super(HumbleDoubleSpinBox, self).focusInEvent(event)

        def focusOutEvent(self, event):
            self.setFocusPolicy(QtCore.Qt.StrongFocus)
            super(HumbleDoubleSpinBox, self).focusOutEvent(event)

        def wheelEvent(self, event):
            if self.hasFocus():
                return super(HumbleDoubleSpinBox, self).wheelEvent(event)
            else:
                event.ignore()
    return HumbleDoubleSpinBox


QHSpinBox: t_.Type[QSpinBox] = humble(QSpinBox)
QHDoubleSpinBox: t_.Type[QDoubleSpinBox] = humble(QDoubleSpinBox)
QHComboBox: t_.Type[QComboBox] = humble(QComboBox)


class VerticallyCompressedWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setLayout(QVBoxLayout())
        self._contentsFrame = QFrame(parent=self)
        spacer = QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.layout().addWidget(self._contentsFrame)
        self.layout().addItem(spacer)
        self.layout = self._layout # override methods
        self.setLayout = self._setLayout

    def _layout(self) -> QLayout:
        return self._contentsFrame.layout()

    def _setLayout(self, layout: QLayout):
        self._contentsFrame.setLayout(layout)


# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.
import logging
import os
import re

from PyQt5 import QtCore
from PyQt5.QtCore import QPoint
from PyQt5.QtGui import QFontMetrics
from PyQt5.QtWidgets import QDockWidget, QWidget, QVBoxLayout, QComboBox, QLineEdit, QGridLayout, QSplitter, \
    QSizePolicy, QMessageBox, QPushButton, QMenu, QAction
import pwspy.dataTypes as pwsdt
from pwspy_gui.PWSAnalysisApp._dockWidgets.CellSelectorDock.widgets import ReferencesTableItem
from .widgets import CellTableWidgetItem, CellTableWidget, ReferencesTable
from pwspy_gui.PWSAnalysisApp.componentInterfaces import CellSelector, ROIManager
from pwspy_gui.PWSAnalysisApp.pluginInterfaces import CellSelectorPluginSupport
from ...sharedWidgets import ScrollableMessageBox
import typing as t_


class CellSelectorDock(CellSelector, QDockWidget):
    """This dockwidget is used by the user to select which cells they want to act upon (run an analysis, plot, etc.)"""

    # noinspection PyUnresolvedReferences
    def __init__(self, parent: QWidget, roiManager: ROIManager):
        super().__init__("Cell Selector", parent=parent)
        self._pluginSupport = CellSelectorPluginSupport(self, self)
        self._roiManager = roiManager
        self._roiManager.roiCreated.connect(lambda roiFile: self.refreshCellItems([roiFile.acquisition]))
        self._roiManager.roiRemoved.connect(lambda roiFile: self.refreshCellItems([roiFile.acquisition]))
        self.setStyleSheet("QDockWidget > QWidget { border: 1px solid lightgray; }")
        self.setObjectName('CellSelectorDock')  # needed for restore state to work
        self._widget = QWidget(self)
        layout = QVBoxLayout()
        addedColumns = []
        for plugin in self._pluginSupport.getPlugins():
            addedColumns += plugin.additionalColumnNames()
        self._tableWidget = CellTableWidget(self._widget, addedColumns)
        self._selectionChangeDebounce = QtCore.QTimer()  # This timer prevents the selectionChanged signal from firing too rapidly.
        self._selectionChangeDebounce.setInterval(500)
        self._selectionChangeDebounce.setSingleShot(True)
        self._selectionChangeDebounce.timeout.connect(lambda: self._pluginSupport.notifyCellSelectionChanged(self.getSelectedCellMetas()))
        self._tableWidget.itemSelectionChanged.connect(self._selectionChangeDebounce.start)

        self._refTableWidget = ReferencesTable(self._widget, self._tableWidget)
        self._refSelectionChangeDebounce = QtCore.QTimer()
        self._refSelectionChangeDebounce.setInterval(500)
        self._refSelectionChangeDebounce.setSingleShot(True)
        self._refSelectionChangeDebounce.timeout.connect(lambda: self._pluginSupport.notifyReferenceSelectionChanged(self.getSelectedReferenceMeta()))
        self._refTableWidget.itemSelectionChanged.connect(self._refSelectionChangeDebounce.start)

        self._bottomBar = QWidget(self._widget)

        self._pathFilter = QComboBox(self._bottomBar)
        self._pathFilter.setEditable(True)

        self._expressionFilter = QLineEdit(self._bottomBar)
        description = "Python boolean expression.\n\tCell#: {num},\n\tAnalysis names: {analyses},\n\tROI names: {roiNames},\n\tRoi numbers: {roiNums},\n\tID tag: {idTag}.\nE.G. `{num} > 5 and 'nucleus' in {roiNames}`"
        self._expressionFilter.setPlaceholderText(description.replace('\n', '').replace('\t', ''))  #Strip out the white space
        self._expressionFilter.setToolTip(description)
        self._expressionFilter.returnPressed.connect(self._executeFilter)

        self._pluginsButton = QPushButton("Tools", self)
        self._pluginsButton.released.connect(self._showPluginMenu)

        _ = QGridLayout()
        _.addWidget(self._pathFilter, 0, 0, 1, 1)
        _.addWidget(self._expressionFilter, 0, 1, 1, 1)
        _.addWidget(self._pluginsButton, 0, 2, 1, 1)
        self._bottomBar.setLayout(_)
        _ = QSplitter()
        _.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        _.addWidget(self._tableWidget)
        _.addWidget(self._refTableWidget)
        _.setCollapsible(0, False)
        _.setCollapsible(1, False)
        _.setSizes([300, 100])
        _.setStretchFactor(1, 0); _.setStretchFactor(0, 1)  # Make the references column so it doesn't resize on stretching.
        layout.addWidget(_)
        layout.addWidget(self._bottomBar)
        self._widget.setLayout(layout)
        self.setWidget(self._widget)

    def _showPluginMenu(self):
        menu = QMenu("plugin menu", self)
        actions = []
        for plugin in self._pluginSupport.getPlugins():
            action = QAction(plugin.getName())
            action.triggered.connect(lambda checked, p=plugin: self._pluginSupport.notifyPluginSelected(p))
            menu.addAction(action)
            actions.append(action)  # Without this the actions get deleted before the menu is shown.
        menu.exec(self._pluginsButton.mapToGlobal(QPoint(0, self._pluginsButton.height())))

    def _addCells(self, acquisitions: t_.List[pwsdt.Acquisition], workingDir: str):
        workingDir = str(workingDir)  # This prevents problems if we pass a Path from pathlib instead.
        cellItems = {}
        for acq in acquisitions:
            addedWidgets = []
            for plugin in self._pluginSupport.getPlugins():
                addedWidgets += plugin.getTableWidgets(acq)
            cellItems[acq] = CellTableWidgetItem(acq, os.path.split(acq.filePath)[0][len(workingDir) + 1:],
                                        int(acq.filePath.split('Cell')[-1]),  additionalWidgets=addedWidgets)
        refItems = [i for i in cellItems.values() if i.isReference()]
        if len(refItems) > 0:
            self._refTableWidget.updateReferences(True, refItems)
        self._tableWidget.addCellItems(cellItems)

    def _clearCells(self):  #This is used publically, probably shouldn't be.
        self._cells = []
        self._tableWidget.clearCellItems()

    def _updateFilters(self):
        try:
            self._pathFilter.currentIndexChanged.disconnect()
        except:
            pass
        self._pathFilter.clear()
        self._pathFilter.addItem('.*')
        paths = []
        for i in self._tableWidget.cellItems.values():
            paths.append(i.path)
        paths = set(paths)  # Get rid of duplicates
        self._pathFilter.addItems(paths)
        maxWidth = max([self._pathFilter.view().fontMetrics().width(text) for text in paths])  # The width needed to show the longest path in the popup.
        self._pathFilter.view().setMinimumWidth(maxWidth)
        self._pathFilter.currentIndexChanged.connect(self._executeFilter)  # reconnect

    def _executeFilter(self): #TODO the filter should also hide the reference items. this will require some changes ot the referece item table code.
        path = self._pathFilter.currentText()
        path = path.replace('\\', '\\\\')
        for item in self._tableWidget.cellItems.values():
            text = item.path.replace(r'\\', r'\\\\')
            try:
                match = re.match(path, text)
            except re.error:
                QMessageBox.information(self, 'Hmm', f'{path} is not a valid regex expression.')
                return
            expr = self._expressionFilter.text()
            if expr.strip() != '':
                analyses = []
                if item.Acquisition.pws:
                    analyses += item.Acquisition.pws.getAnalyses()
                if item.Acquisition.dynamics:
                    analyses += item.Acquisition.dynamics.getAnalyses()
                try:
                    roiNames, roiNums, fformats = zip(*item.Acquisition.getRois())
                except ValueError as ve:  # If an acquisition has no ROIs then the 0 items returned can't be unpacked to 3 variables
                    roiNames = []; roiNums = []; fformats = []
                try:
                    ret = bool(eval(expr.format(num=item.num,
                                                analyses=analyses,
                                                roiNames=roiNames,
                                                roiNums=roiNums,
                                                idTag=item.Acquisition.idTag)))
                except Exception as e:
                    QMessageBox.information(self, 'Hmm', f'{expr} is not a valid boolean expression.')
                    logging.getLogger(__name__).exception(e)
                    return
            else:
                ret = True
            if match and ret:
                self._tableWidget.setRowHidden(item.row, False)
            else:
                self._tableWidget.setRowHidden(item.row, True)

    def close(self):
        """This makes sure the application metadata is saved."""
        self._clearCells()

    def getRoiManager(self) -> ROIManager:
        return self._roiManager

    def loadNewCells(self, fileNames: t_.List[str], workingDir: str):
        self._clearCells()
        acqs = []
        errs = []
        for f in fileNames:
            try:
                acqs.append(pwsdt.Acquisition(f))
            except OSError as e:
                logger = logging.getLogger(__name__)
                errs.append(f)
                logger.warning(f"Failed to load {f}")
                logger.exception(e)
                continue
        if len(errs) > 0:
            ScrollableMessageBox.information(self, "Load Errors", f"The following images failed to load:\n {', '.join(errs)}")
        self._pluginSupport.notifyNewCellsLoaded(acqs)
        self._addCells(acqs, workingDir)
        self._updateFilters()

    def getSelectedCellMetas(self) -> t_.Tuple[pwsdt.Acquisition]:
        return tuple(i.Acquisition for i in self._tableWidget.selectedCellItems)

    def getAllCellMetas(self) -> t_.Tuple[pwsdt.Acquisition]:
        return tuple(self._tableWidget.cellItems.keys())

    def getSelectedReferenceMeta(self) -> t_.Optional[pwsdt.Acquisition]:
        return self._refTableWidget.selectedReferenceMeta

    def setSelectedCells(self, cells: t_.List[pwsdt.Acquisition]):
        idTags = [i.idTag for i in cells]
        for item in self._tableWidget.cellItems.values():
            if item.Acquisition.idTag in idTags:
                item.setSelected(True)
            else:
                item.setSelected(False)

    def setSelectedReference(self, ref: pwsdt.Acquisition):
        idTag = ref.idTag
        for i in range(self._refTableWidget.rowCount()):
            refitem: ReferencesTableItem = self._refTableWidget.item(i, 0)
            if refitem.item.Acquisition.idTag == idTag:
                refitem.setSelected(True)
            else:
                refitem.setSelected(False)

    def setHighlightedCells(self, cells: t_.List[pwsdt.Acquisition]):
        idTags = [i.idTag for i in cells]
        for item in self._tableWidget.cellItems.values():
            if item.Acquisition.idTag in idTags:
                item.setHighlighted(True)
            else:
                item.setHighlighted(False)

    def setHighlightedReference(self, ref: pwsdt.Acquisition):
        idTag = ref.idTag
        for i in range(self._refTableWidget.rowCount()):
            refitem: ReferencesTableItem = self._refTableWidget.item(i, 0)
            if refitem.item.Acquisition.idTag == idTag:
                refitem.setHighlighted(True)
            else:
                refitem.setHighlighted(False)

    def refreshCellItems(self, cells: t_.List[pwsdt.Acquisition] = None):
        """`Cells` indicates which cells need refreshing. If cells is None then all cells will be refreshed."""
        self._tableWidget.refreshCellItems(cells=cells)



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
import re
import traceback

import numpy as np
from PyQt5 import QtCore
from PyQt5.QtGui import QCursor, QValidator
from PyQt5.QtWidgets import QWidget, QDialog, QVBoxLayout, QDoubleSpinBox, QPushButton, QLabel, QGridLayout, QComboBox, \
    QAction, QMenu, QApplication, QCheckBox
from matplotlib.backends.backend_qt5 import NavigationToolbar2QT as NavigationToolbar
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure

from pwspy_gui.sharedWidgets.rangeSlider import QRangeSlider
import typing
if typing.TYPE_CHECKING:
    from pwspy.dataTypes import Acquisition


class BigPlot(QWidget):
    def __init__(self, data: np.ndarray, parent=None):
        """A widget that displays an image."""
        super().__init__(parent=parent)
        layout = QGridLayout()
        self.fig = Figure()
        self.ax = self.fig.add_subplot(1, 1, 1)
        self.ax.get_yaxis().set_visible(False)
        self.ax.get_xaxis().set_visible(False)
        self.im = self.ax.imshow(data, cmap='gray')
        self.fig.colorbar(self.im, ax=self.ax)
        self.canvas = FigureCanvasQTAgg(self.fig)
        self.canvas.setFocusPolicy(QtCore.Qt.ClickFocus)
        self.canvas.setFocus()
        self.fig.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)
        self.slider = QRangeSlider(self)
        self.slider.endValueChanged.connect(self._climImage)
        self.slider.startValueChanged.connect(self._climImage)
        self.autoDlg = SaturationDialog(self)
        self.autoDlg.accepted.connect(self._setSaturation)
        self.rangeDlg = RangeDialog(self)
        self.rangeDlg.accepted.connect(self._setRange)
        self.saturationButton = QPushButton("Auto")
        self.saturationButton.released.connect(self.autoDlg.show)
        self.manualRangeButton = QPushButton("Range")
        self.manualRangeButton.released.connect(self.rangeDlg.show)
        self.cmapCombo = QComboBox(self)
        self.cmapCombo.addItems(['gray', 'jet', 'plasma', 'Reds'])
        self.cmapCombo.currentTextChanged.connect(self._changeCmap)

        layout.addWidget(self.canvas, 0, 0, 8, 8)
        layout.addWidget(QLabel("Color Range"), 9, 0, 1, 1)
        layout.addWidget(self.slider, 9, 1, 1, 4)
        layout.addWidget(self.saturationButton, 9, 6, 1, 1)
        layout.addWidget(self.manualRangeButton, 9, 7, 1, 1)
        layout.addWidget(NavigationToolbar(self.canvas, self), 10, 0, 1, 4)

        layout.addWidget(QLabel("Color Map"), 10, 6, 1, 1, alignment=QtCore.Qt.AlignRight)
        layout.addWidget(self.cmapCombo, 10, 7, 1, 1)
        layout.setRowStretch(0, 1)  # This causes the plot to take up all the space that isn't needed by the other widgets.
        l = QVBoxLayout()
        l.addLayout(layout)
        self.setLayout(l)

        self.setImageData(data)
        self._setSaturation()


    def setImageData(self, data: np.ndarray):
        self.data = data
        self.im.set_data(data)
        self.slider.setMax(np.nanmax(self.data))
        self.slider.setMin(np.nanmin(self.data))
        if self.autoDlg.autoSaturateCheckBox.isChecked():
            self._setSaturation()
        self.canvas.draw_idle()

    def _setSaturation(self):
        percentage = self.autoDlg.value
        m = np.nanpercentile(self.data, percentage)
        M = np.nanpercentile(self.data, 100 - percentage)
        self.slider.setStart(m)
        self.slider.setEnd(M)

    def _setRange(self):
        self.slider.setStart(self.rangeDlg.minimum)
        self.slider.setEnd(self.rangeDlg.maximum)

    def _climImage(self):
        self.im.set_clim((self.slider.start(), self.slider.end()))
        self.canvas.draw_idle()

    def _changeCmap(self, cMap: str):
        self.im.set_cmap(cMap)
        self.canvas.draw_idle()


class SaturationDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent=parent, flags=QtCore.Qt.FramelessWindowHint)
        self.setModal(True)
        l = QGridLayout()
        self.numBox = QDoubleSpinBox()
        self.numBox.setValue(0.1)
        self.numBox.setMinimum(0)
        self.numBox.setSingleStep(0.1)
        self.autoSaturateCheckBox = QCheckBox("Update contrast when image is changed", self)
        self.autoSaturateCheckBox.setChecked(True)
        self.okButton = QPushButton("Ok")
        self.okButton.released.connect(self.accept)
        l.addWidget(QLabel("Saturation %"), 0, 0)
        l.addWidget(self.numBox, 0, 1)
        l.addWidget(self.autoSaturateCheckBox, 1, 0, 1, 2)
        l.addWidget(self.okButton, 2, 0, 1, 2)
        self.setLayout(l)

    @property
    def value(self):
        return self.numBox.value()


class RangeDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent=parent, flags=QtCore.Qt.FramelessWindowHint)
        self.setModal(True)
        l = QGridLayout()
        self.minBox = QDoubleSpinBox()

        self.maxBox = QDoubleSpinBox()
        self.okButton = QPushButton("Ok")
        self.okButton.released.connect(self.accept)
        l.addWidget(QLabel("Min"), 0, 0, 1, 1)
        l.addWidget(QLabel("Max"), 0, 1, 1, 1)
        l.addWidget(self.minBox, 1, 0, 1, 1)
        l.addWidget(self.maxBox, 1, 1, 1, 1)
        l.addWidget(self.okButton, 2, 1, 1, 1)
        self.setLayout(l)

    def show(self):
        for b in [self.minBox, self.maxBox]:
            b.setMaximum(self.parent().slider.max())
            b.setMinimum(self.parent().slider.min())
        self.minBox.setValue(self.parent().slider.start())
        self.maxBox.setValue(self.parent().slider.end())
        super().show()

    @property
    def minimum(self): return self.minBox.value()

    @property
    def maximum(self): return self.maxBox.value()


if __name__ == '__main__':
    fPath = r'C:\Users\nicke\Desktop\demo\toast\t\Cell1'
    from pwspy.dataTypes import Acquisition
    acq = Acquisition(fPath)
    import sys
    app = QApplication(sys.argv)
    b = BigPlot(acq.dynamics.getThumbnail())
    b.show()
    sys.exit(app.exec())



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
import traceback

import logging
from PyQt5.QtCore import QThread
from PyQt5.QtWidgets import QMessageBox, QMainWindow
from PyQt5 import QtCore
from pwspy_gui.sharedWidgets.dialogs import BusyDialog
from pwspy_gui.PWSAnalysisApp._taskManagers.analysisManager import safeCallback
import re
from pwspy_gui.PWSAnalysisApp.utilities.conglomeratedAnalysis import ConglomerateCompiler, ConglomerateAnalysisResults
import typing
if typing.TYPE_CHECKING:
    from typing import Tuple, List, Optional
    from pwspy.dataTypes import Acquisition
    from pwspy_gui.PWSAnalysisApp.utilities.conglomeratedAnalysis import ConglomerateCompilerSettings, ConglomerateCompilerResults
    from pwspy.analysis.warnings import AnalysisWarning


class CompilationManager(QtCore.QObject):
    compilationDone = QtCore.pyqtSignal(list)

    def __init__(self, window: QMainWindow):
        super().__init__()
        self.window = window

    @safeCallback
    def run(self) -> List[Tuple[Acquisition, List[Tuple[ConglomerateCompilerResults, Optional[List[AnalysisWarning]]]]]]:
        roiName: str = self.window.resultsTable.getRoiName()
        analysisName: str = self.window.resultsTable.getAnalysisName()
        settings: ConglomerateCompilerSettings = self.window.resultsTable.getSettings()
        cellMetas: List[Acquisition] = self.window.cellSelector.getSelectedCellMetas()
        if len(cellMetas) == 0:
            QMessageBox.information(self.window, "What?", "Please select at least one cell.")
            return None
        compiler = ConglomerateCompiler(settings)
        b = BusyDialog(self.window, "Processing. Please Wait...")
        t = self.CompilationThread(cellMetas, compiler, roiName, analysisName)
        t.finished.connect(b.accept)
        t.errorOccurred.connect(lambda e: QMessageBox.information(self.window, 'Uh Oh', str(e)))
        t.start()
        b.exec()
        results = t.result
        if results is not None: #If None then an error must have occurred
            self.compilationDone.emit(results)
            return results
        else:
            return None

    class CompilationThread(QThread):
        errorOccurred = QtCore.pyqtSignal(Exception)

        def __init__(self, cellMetas: List[Acquisition], compiler: ConglomerateCompiler, roiNamePattern: str, analysisNamePattern: str):
            super().__init__()
            self.cellMetas = cellMetas
            self.roiNamePattern = roiNamePattern
            self.analysisNamePattern = analysisNamePattern
            self.compiler = compiler
            self.result = None

        def run(self):
            try:
                self.result = []
                for acq in self.cellMetas:
                    self.result.append(self._process(acq, self.compiler, self.roiNamePattern, self.analysisNamePattern)) # A list of Tuples. each tuple containing a list of warnings and the Acquisition to go with it.
            except Exception as e:
                logger = logging.getLogger(__name__)
                logger.warning("Compilation error:")
                logger.exception(e)
                self.errorOccurred.emit(e)

        @staticmethod
        def _process(acq: Acquisition, compiler: ConglomerateCompiler, roiNamePattern: str, analysisNamePattern: str) -> Tuple[Acquisition, List[Tuple[ConglomerateCompilerResults, List[AnalysisWarning]]]]:
            rois = [acq.loadRoi(name, num, fformat) for name, num, fformat in acq.getRois() if re.match(roiNamePattern, name)]
            pwsAnalysisResults = [acq.pws.loadAnalysis(name) for name in acq.pws.getAnalyses() if re.match(analysisNamePattern, name)] if acq.pws is not None else []
            dynamicAnalysisResults = [acq.dynamics.loadAnalysis(name) for name in acq.dynamics.getAnalyses() if re.match(analysisNamePattern, name)] if acq.dynamics is not None else []
            conglomeratedAnalysisResults = []
            for pws in pwsAnalysisResults:  # Find the analyses with matching names and pair them.
                for dyn in dynamicAnalysisResults:
                    if pws.analysisName == dyn.analysisName:
                        conglomeratedAnalysisResults.append(ConglomerateAnalysisResults(pws, dyn))
                        pwsAnalysisResults.remove(pws) #Once an analysis has been paired remove it from the list of analyses
                        dynamicAnalysisResults.remove(dyn)
            conglomeratedAnalysisResults += [ConglomerateAnalysisResults(pws, None) for pws in pwsAnalysisResults] #Any remaining analyses couldn't be paired. Just add them on their own.
            conglomeratedAnalysisResults += [ConglomerateAnalysisResults(None, dyn) for dyn in dynamicAnalysisResults]
            ret = []
            for analysisResult in conglomeratedAnalysisResults:
                for roi in rois:
                    cResults, warnings = compiler.run(analysisResult, roi)
                    ret.append((cResults, warnings))
            return acq, ret



from __future__ import annotations
import matplotlib
from PyQt5 import QtCore
from PyQt5.QtWidgets import QDialog, QWidget, QVBoxLayout, QDoubleSpinBox, QSpinBox, QPushButton, QGridLayout, QLabel, \
    QHBoxLayout, QMessageBox
from matplotlib.backends.backend_qt5 import NavigationToolbar2QT
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
import typing as t_
from pwspy.utility.plotting import roiColor
if t_.TYPE_CHECKING:
    from pwspy_gui.PWSAnalysisApp.sharedWidgets.plotting import RoiPlot


class SinCityDlg(QDialog):
    def __init__(self, parentRoiPlot: RoiPlot, parent: QWidget = None):
        super().__init__(parent=parent)
        self.setWindowFlags(QtCore.Qt.Window | QtCore.Qt.WindowMinMaxButtonsHint)
        # self.setWindowFlags(QtCore.Qt.Window | QtCore.Qt.WindowTitleHint | QtCore.Qt.CustomizeWindowHint)  # Get rid of the close button. this is handled by the selector widget active status
        self.setWindowTitle("Sin City Image Export")
        self.setModal(False)

        self.parentRoiPlot = parentRoiPlot
        self.cachedImage = None

        self.fig, self.ax = matplotlib.pyplot.subplots()
        c = FigureCanvasQTAgg(self.fig)
        self.plotWidg = QWidget(self)
        self.plotWidg.setLayout(QVBoxLayout())
        self.plotWidg.layout().addWidget(c)
        self.plotWidg.layout().addWidget(NavigationToolbar2QT(c, self))

        self.ax.xaxis.set_visible(False)
        self.ax.yaxis.set_visible(False)
        self.im = self.ax.imshow(self.parentRoiPlot.getImageData())

        self._paintDebounce = QtCore.QTimer()  # This timer prevents the selectionChanged signal from firing too rapidly.
        self._paintDebounce.setInterval(200)
        self._paintDebounce.setSingleShot(True)
        self._paintDebounce.timeout.connect(self.paint)

        self.vmin = QDoubleSpinBox(self)
        self.vmin.setValue(0)
        self.vmin.setDecimals(3)
        self.vmin.setMaximum(10000)
        self.vmin.setSingleStep(0.001)

        def vminChanged(val):
            self.stale = True
            self._paintDebounce.start()
        self.vmin.valueChanged.connect(vminChanged)

        self.vmax = QDoubleSpinBox(self)
        self.vmax.setValue(.1)
        self.vmax.setDecimals(3)
        self.vmax.setMaximum(10000)
        self.vmax.setSingleStep(0.001)

        def vmaxChanged(val):
            self.stale = True
            self._paintDebounce.start()
        self.vmax.valueChanged.connect(vmaxChanged)

        self.scaleBg = QDoubleSpinBox(self)
        self.scaleBg.setValue(.33)
        self.scaleBg.setMinimum(0)
        self.scaleBg.setDecimals(2)
        self.scaleBg.setMaximum(3)
        self.scaleBg.setSingleStep(0.01)

        def scaleBgChanged(val):
            self.stale = True
            self._paintDebounce.start()
        self.scaleBg.valueChanged.connect(scaleBgChanged)

        self.hue = QDoubleSpinBox(self)
        self.hue.setMinimum(0)
        self.hue.setMaximum(1)
        self.hue.setValue(0)
        self.hue.setSingleStep(0.05)

        def hueRangeChanged(val):
            self.stale = True
            self._paintDebounce.start()
        self.hue.valueChanged.connect(hueRangeChanged)

        self.exp = QDoubleSpinBox(self)
        self.exp.setMinimum(0.5)
        self.exp.setMaximum(3)
        self.exp.setValue(1)
        self.exp.setSingleStep(0.05)

        def expRangeChanged(val):
            self.stale = True
            self._paintDebounce.start()
        self.exp.valueChanged.connect(expRangeChanged)

        self.scaleBar = QSpinBox(self)
        self.scaleBar.setValue(0)
        self.scaleBar.setMaximum(10000)

        def scaleBarChanged(val):
            self.stale = True
            self._paintDebounce.start()
        self.scaleBar.valueChanged.connect(scaleBarChanged)

        self.refreshButton = QPushButton("Refresh", self)

        def refreshAction():
            self.stale = True  # Force a full refresh
            self.paint()

        self.refreshButton.released.connect(refreshAction)

        l = QGridLayout()
        l.addWidget(QLabel("Minimum Value", self), 0, 0)
        l.addWidget(self.vmin, 0, 1)
        l.addWidget(QLabel("Maximum Value", self), 1, 0)
        l.addWidget(self.vmax, 1, 1)
        l.addWidget(QLabel("Scale Background", self), 2, 0)
        l.addWidget(self.scaleBg, 2, 1)
        l.addWidget(QLabel("Hue", self), 3, 0)
        l.addWidget(self.hue, 3, 1)
        l.addWidget(QLabel("Exponent", self), 4, 0)
        l.addWidget(self.exp, 4, 1)
        l.addWidget(QLabel("Scale Bar", self), 5, 0)
        l.addWidget(self.scaleBar, 5, 1)
        l.addWidget(self.refreshButton, 6, 0)
        layout = QHBoxLayout()
        layout.addLayout(l)
        layout.addWidget(self.plotWidg)
        self.setLayout(layout)

        self.stale = True
        self.paint()

    def paint(self):
        """Refresh the recommended regions. If stale is false then just repaint the cached regions without recalculating."""
        if self.parentRoiPlot.getImageData() is not self.cachedImage:  # The image has been changed.
            self.cachedImage = self.parentRoiPlot.getImageData()
            self.stale = True
        if self.stale:
            try:
                rois = [parm.roiFile for parm in self.parentRoiPlot.rois]
                data = roiColor(self.parentRoiPlot.getImageData(), [r.getRoi() for r in rois], self.vmin.value(), self.vmax.value(), self.scaleBg.value(), hue=self.hue.value(), exponent=self.exp.value(), numScaleBarPix=self.scaleBar.value())
                self.im.set_data(data)
                self.fig.canvas.draw_idle()
                self.stale = False
            except Exception as e:
                msg = QMessageBox.information(self, "Error", f"Warning: Sin City export failed with error: {str(e)}")
                return


# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from PyQt5.QtWidgets import QDialog, QWidget, QTextEdit, QGridLayout, QLabel, QPushButton


class IndexInfoForm(QDialog):
    """This dialog is used to collect a description about the circumstances surrounding the collection of data for an Extra Reflectance calibration.
    This description is ultimately saved to an index which is acts as a sort of database for all of the calibrations."""
    def __init__(self, name, idTag, parent: QWidget = None):
        super().__init__(parent=parent)
        self.description = None
        self.setModal(False)
        self.setWindowTitle(name)
        self.descriptionEdit = QTextEdit()
        self.okButton = QPushButton('OK!')
        self.okButton.released.connect(self.accept)
        layout = QGridLayout()
        layout.addWidget(QLabel("Description:", parent=self), 0, 0, 1, 1)
        layout.addWidget(self.descriptionEdit, 1, 0, 4, 4)
        layout.addWidget(QLabel(f"Name: {name}", self), 5, 0, 1, 1)
        layout.addWidget(QLabel(f"idTag: {idTag}", self), 5, 1, 1, 1)
        layout.addWidget(self.okButton, 5, 3, 1, 1)
        self.setLayout(layout)

    def accept(self) -> None:
        self.description = self.descriptionEdit.toPlainText()
        super().accept()



from __future__ import annotations
import typing
from PyQt5 import QtCore
from PyQt5.QtCore import QSize, pyqtSignal
from PyQt5.QtGui import QTextDocument, QAbstractTextDocumentLayout, QPalette
from PyQt5.QtWidgets import QStyledItemDelegate, QWidget, QStyleOptionViewItem, QStyle, QGridLayout, QTableWidget, \
    QTableWidgetItem, QAbstractItemView, QSizePolicy, QApplication
from pwspy.utility.acquisition.steps import SequencerStep, IterableSequencerStep, StepTypeNames, ContainerStep


class IterationRangeEditor(QWidget):
    """
    This widget appears as a QTableWidget that allows the user to select which iterations should be selected for acquisition steps that have multiple
    iterations such as time series or multiple position steps.

    Args:
        parent: The `parent` QWidget of this widget.
    """
    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.setAutoFillBackground(True)  # Prevents us from getting double vision with the painted version of the widget behind.
        self._coordTable = QTableWidget(self)
        self._coordTable.setEditTriggers(QAbstractItemView.NoEditTriggers)

        # This makes the items stay looking selected even when the table is inactive
        self._coordTable.setStyleSheet("""QTableWidget::item:active {   
                                        selection-background-color: darkblue;
                                        selection-color: white;}
                                        QTableWidget::item:inactive {
                                        selection-background-color: darkblue;
                                        selection-color: white;}""")

        self._coordTable.setRowCount(1)
        self._coordTable.horizontalHeader().setVisible(False)
        self._coordTable.verticalHeader().setVisible(False)
        self._coordTable.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)  # Always show scrollbar
        self._coordTable.horizontalScrollBar().setStyleSheet(
            "QScrollBar:horizontal { height: 15px; }"  # For some reason setting the height directly doesn't work.
        )
        self._coordTable.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)  # Smooth scrolling
        self._coordTable.sizePolicy().setHorizontalPolicy(QSizePolicy.Expanding)

        self._requiredHeight = self._coordTable.horizontalScrollBar().height() + self._coordTable.rowHeight(0)

        l = QGridLayout()
        l.setContentsMargins(0, 0, 0, 0)
        l.setVerticalSpacing(0)
        l.addWidget(self._coordTable, 0, 0, 1, 1, alignment=QtCore.Qt.AlignLeft)
        l.setRowStretch(1, 1)  # Allow the actual wigdets to sit at the top of the layout without getting stretched out.
        self.setLayout(l)

    def setFromStep(self, step: IterableSequencerStep) -> None:
        """
        Update the appearance and labels of this widget based on the `Step` provided as input.

        Args:
            step: An iterable step from the sequencer.
        """
        self._coordTable.setColumnCount(step.stepIterations())
        for i in range(step.stepIterations()):
            self._coordTable.setItem(0, i, QTableWidgetItem(step.getIterationName(i)))
        self._coordTable.resizeColumnsToContents()
        self._coordTable.resizeRowsToContents()

        h2 = self._coordTable.rowHeight(0)
        h1 = self._coordTable.horizontalScrollBar().sizeHint().height()  # Just getting the current height doesn't work for some reason.
        self._coordTable.setMaximumHeight(h1+h2)
        self._coordTable.setMinimumHeight(h1+h2)

        w = 0
        for i in range(step.stepIterations()):
            w += self._coordTable.columnWidth(i)
        self._coordTable.setMaximumWidth(w+2)  # The plus 2 gets rid of the scrollbar.

        #Make the selection match
        for i in range(step.stepIterations()):
            selectedIterations: typing.Tuple[int, typing.Sequence[int]] = step.data(QtCore.Qt.EditRole)
            if selectedIterations is None: selectedIterations = (step.id, [])  # no iterations selected, should be treated the same as having all iterations selected
            if i in selectedIterations[1]:
                sel = True
            else:
                sel = False
            self._coordTable.item(0, i).setSelected(sel)

    def getSelection(self) -> typing.List[int]:
        """
        Return a list of the column indices that are selected.
        """
        return [i.column() for i in self._coordTable.selectedIndexes()]

    def sizeHint(self) -> QtCore.QSize:
        """
        Overrides the default QT function to help with proper sizing.
        """
        h2 = self._coordTable.height()
        return QtCore.QSize(1, h2)  # Width doesn't seem to matter


class HTMLDelegate(QStyledItemDelegate):
    """
    Overrides the default QT delegate that displays and edits data in a QT `Model`.
    Allows rendering of HTML strings.
    """

    def paint(self, painter, option, index):
        self.initStyleOption(option, index)
        style = option.widget.style() if option.widget else QApplication.style()
        doc = QTextDocument()
        doc.setHtml(option.text)

        # Painting item without text
        option.text = ""
        style.drawControl(QStyle.CE_ItemViewItem, option, painter)

        ctx = QAbstractTextDocumentLayout.PaintContext()

        # Highlighting text if item is selected
        if (option.state & QStyle.State_Selected):
            ctx.palette.setColor(QPalette.Text, option.palette.color(QPalette.Active, QPalette.HighlightedText))

        textRect = style.subElementRect(QStyle.SE_ItemViewItemText, option)
        painter.save()
        painter.translate(textRect.topLeft())
        painter.setClipRect(textRect.translated(-textRect.topLeft()))
        doc.documentLayout().draw(painter, ctx)
        painter.restore()

    def sizeHint(self, option, index):
        self.initStyleOption(option, index)
        doc = QTextDocument()
        doc.setHtml(option.text)
        doc.setTextWidth(option.rect.width())
        return QSize(doc.idealWidth(), doc.size().height())


class IterationRangeDelegate(HTMLDelegate):
    """
    An HTMLDelegate that is specific for the display of a sequencer `Step` that has multiple iterations.

    Args:
        parent: The parent QWidget of this object.

    Attributes:
        editingFinished: This Qt signal is fired when the delegate finishes edited it's data in the model.
    """

    editingFinished = pyqtSignal()

    def __init__(self, parent: QWidget = None):
        super().__init__(parent=parent)
        self._editing = False
        self.editor = None

    def createEditor(self, parent: QWidget, option: QStyleOptionViewItem, index: QtCore.QModelIndex) -> QWidget:
        if isinstance(index.internalPointer(), IterableSequencerStep):
            widg = IterationRangeEditor(parent)
            widg.setFromStep(index.internalPointer())
            widg.resize(option.rect.size())
            self.editor: IterationRangeEditor = widg
            return widg
        else:
            return None  # Don't allow handling other types of values. # super().createEditor(parent, option, index)

    def setModelData(self, editor: IterationRangeEditor, model: QtCore.QAbstractItemModel, index: QtCore.QModelIndex) -> None:
        step: SequencerStep = index.internalPointer()
        if isinstance(step, IterableSequencerStep):
            coordRange = (step.id, editor.getSelection())
            step.setData(QtCore.Qt.EditRole, coordRange)
        self._editing = None
        self.sizeHintChanged.emit(index)
        self.editingFinished.emit()

    def editorEvent(self, event: QtCore.QEvent, model: QtCore.QAbstractItemModel, option: 'QStyleOptionViewItem', index: QtCore.QModelIndex) -> bool:
        self._editing = index
        self.sizeHintChanged.emit(index)
        return super().editorEvent(event, model, option, index)

    def sizeHint(self, option: QStyleOptionViewItem, index: QtCore.QModelIndex) -> QtCore.QSize:
        if isinstance(index.internalPointer(), IterableSequencerStep) and self._editing == index:
            self.initStyleOption(option, index)
            editor = self.createEditor(None, option, index)
            s = editor.sizeHint()
            return s
        else:
            return super().sizeHint(option, index)

    def displayText(self, value: typing.Any, locale: QtCore.QLocale) -> str:
        if isinstance(value, IterableSequencerStep):
            itRangeCoord: typing.Tuple[int, typing.Sequence[int]] = value.data(QtCore.Qt.EditRole)
            if itRangeCoord is None:
                selectedIterations = []  # TODO this happens for steps where we haven't assigned any iterations (non iterable step types). Should we have a `None` here?
            else:
                selectedIterations = itRangeCoord[1]
            if len(selectedIterations) == 0 or len(selectedIterations) == value.stepIterations():
                s = ": all coords"
            else:
                numCoords = len(selectedIterations)
                s = f": {numCoords} coord"
                if numCoords > 1:
                    s += 's'
            return "<html>" + StepTypeNames[value.stepType] + "<b>" + s + "</b>" + "</html>"
        if isinstance(value, ContainerStep):
            return StepTypeNames[value.stepType]  # Just return the name.
        if isinstance(value, SequencerStep):
            return "\u2022" + StepTypeNames[value.stepType]  # For endpoint steps add a bullet
        else:
            super().displayText(value, locale)



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from PyQt5.QtWidgets import (QDockWidget, QWidget, QFrame, QVBoxLayout, QCheckBox, QScrollArea, QPushButton,
                             QGridLayout, QLineEdit, QLabel, QDialog, QTreeWidget, QTreeWidgetItem)
from PyQt5 import QtCore
from pwspy.analysis.compilation import DynamicsCompilerSettings, GenericCompilerSettings, PWSCompilerSettings, \
    PWSRoiCompilationResults
from pwspy_gui.PWSAnalysisApp.utilities.conglomeratedAnalysis import ConglomerateCompilerResults, ConglomerateCompilerSettings
from .widgets import ResultsTable, ResultsTableItem
import typing

from ..._taskManagers.compilationManager import CompilationManager
from ...componentInterfaces import ResultsTableController

if typing.TYPE_CHECKING:
    from pwspy.dataTypes import Acquisition, PwsMetaData
    from pwspy.analysis.pws import PWSAnalysisSettings
    from pwspy.analysis.warnings import AnalysisWarning


class ResultsTableControllerDock(ResultsTableController, QDockWidget):
    def __init__(self, parent: QWidget):
        super().__init__("Results", parent=parent)
        self.setStyleSheet("QDockWidget > QWidget { border: 1px solid lightgray; }")
        self.setObjectName('ResultsTableDock')
        self._widget = QWidget()
        self._widget.setLayout(QGridLayout())
        self._table = ResultsTable()
        checkBoxFrame = QFrame(parent=self)
        checkBoxFrame.setLayout(QVBoxLayout())
        checkBoxFrame.layout().setContentsMargins(1, 1, 1, 1)
        checkBoxFrame.layout().setSpacing(1)
        self._checkBoxes = []
        for i, (name, (default, settingsName, compilerClass, tooltip)) in enumerate(self._table.columns.items()):
            c = QCheckBox(name)
            c.setCheckState(2) if default else c.setCheckState(0)
            c.stateChanged.connect(lambda state, j=i: self._table.setColumnHidden(j, state == 0))
            checkBoxFrame.layout().addWidget(c)
            self._checkBoxes.append(c)
        self._roiNameEdit = QLineEdit('.*', self._widget)
        self._roiNameEdit.setToolTip("ROIs matching this RegEx pattern will be compiled.")
        self._analysisNameEdit = QLineEdit('.*', self._widget)
        self._analysisNameEdit.setToolTip("Analyses matching this RegEx pattern will be compiled.")
        self._compileButton = QPushButton("Compile")

        self._compMan = CompilationManager(self.window())
        self._compileButton.released.connect(self._compMan.run)
        self._compMan.compilationDone.connect(self._handleCompilationResults)

        scroll = QScrollArea()
        scroll.setWidget(checkBoxFrame)
        scroll.verticalScrollBar().setStyleSheet("QScrollBar:horizontal { height: 10px; }")
        scroll.setMaximumWidth(checkBoxFrame.width() + 10)
        scroll.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        scroll.horizontalScrollBar().setEnabled(False)
        sidebar = QWidget()
        l = QGridLayout()
        l.addWidget(scroll, 0, 0, 1, 2)
        l.addWidget(QLabel('Analysis:'), 1, 0, 1, 1)
        l.addWidget(self._analysisNameEdit, 1, 1, 1, 1)
        l.addWidget(QLabel("Roi:"), 2, 0, 1, 1)
        l.addWidget(self._roiNameEdit, 2, 1, 1, 1)
        l.addWidget(self._compileButton, 3, 0, 1, 2)
        sidebar.setLayout(l)
        sidebar.setMaximumWidth(scroll.width()+10)
        self._widget.layout().addWidget(sidebar, 0, 0)
        self._widget.layout().addWidget(self._table, 0, 1)
        self.setWidget(self._widget)

    def addCompilationResult(self, result: ConglomerateCompilerResults, acquisition: Acquisition):
        self._table.addItem(ResultsTableItem(result, acquisition))

    def clearCompilationResults(self):
        self._table.clearCellItems()

    def getSettings(self) -> ConglomerateCompilerSettings:
        pwskwargs = {}
        dynkwargs = {}
        genkwargs = {}
        for checkBox in self._checkBoxes:
            defaultVisible, settingsName, compilerClass, tooltip = self._table.columns[checkBox.text()]
            if settingsName is not None:
                if compilerClass == PWSCompilerSettings:
                    pwskwargs[settingsName] = bool(checkBox.checkState())
                elif compilerClass == DynamicsCompilerSettings:
                    dynkwargs[settingsName] = bool(checkBox.checkState())
                elif compilerClass == GenericCompilerSettings:
                    genkwargs[settingsName] = bool(checkBox.checkState())
                else:
                    raise ValueError("All member of ResultsTable.columns that have a setting name defined must also have an associated compilerClass value")
        pws = PWSCompilerSettings(**pwskwargs)
        dyn = DynamicsCompilerSettings(**dynkwargs)
        gen = GenericCompilerSettings(**genkwargs)
        return ConglomerateCompilerSettings(pws, dyn, gen)

    def getRoiName(self) -> str:
        return self._roiNameEdit.text()

    def getAnalysisName(self) -> str:
        return self._analysisNameEdit.text()

    def _handleCompilationResults(self, inVal: typing.List[typing.Tuple[Acquisition, typing.List[typing.Tuple[ConglomerateCompilerResults, typing.Optional[typing.List[AnalysisWarning]]]]]]):
        #  Display warnings if necessary.
        warningStructure = []
        for acq, roiList in inVal:
            metaWarnings = []
            for result, warnList in roiList:
                if len(warnList) > 0:
                    metaWarnings.append((result, warnList))
            if len(metaWarnings) > 0:
                warningStructure.append((acq.pws, metaWarnings))
        if len(warningStructure) > 0:
            CompilationSummaryDisplay(self.window(), warningStructure)
        #  Display the results on the table
        results = [(acq, result) for acq, roiList in inVal for result, warnings in roiList]
        self.clearCompilationResults()
        [self.addCompilationResult(r, acq) for acq, r in results]


class CompilationSummaryDisplay(QDialog):
    def __init__(self, parent: typing.Optional[QWidget], warnings: typing.List[typing.Tuple[PwsMetaData, typing.List[typing.Tuple[PWSRoiCompilationResults, typing.Optional[typing.List[AnalysisWarning]]]]]], analysisName: str = '', analysisSettings: PWSAnalysisSettings = None):
        super().__init__(parent=parent)
        self.setWindowTitle("Compilation Summary")
        layout = QVBoxLayout()
        self.warningTree = QTreeWidget(self)
        self.warningTree.setHeaderHidden(True)
        layout.addWidget(self.warningTree)
        self.setLayout(layout)
        self._addWarnings(warnings)
        self.show()

    def _addWarnings(self, warnings: typing.List[typing.Tuple[PwsMetaData, typing.List[typing.Tuple[ConglomerateCompilerResults, typing.Optional[typing.List[AnalysisWarning]]]]]]):
        for meta, roiList in warnings:
            item = QTreeWidgetItem(self.warningTree)
            item.setText(0, meta.filePath)
            for roiResult, roiWarnList in roiList:
                if len(roiWarnList) > 0:
                    subItem = QTreeWidgetItem(item)
                    subItem.setText(0, f"{len(roiWarnList)} warnings: {roiResult.generic.roiFile.name} {roiResult.generic.roiFile.number}")
                    for warn in roiWarnList:
                        subItem2 = QTreeWidgetItem(subItem)
                        subItem2.setText(0, warn.shortMsg)
                        subItem2.setToolTip(0, warn.longMsg)

    def clearWarnings(self):
        self.warningTree.clear()



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
import json
import logging
import os
from datetime import datetime
from typing import List, Optional, TextIO

import jsonschema

from pwspy import dateTimeFormat


class ERIndex:
    FILENAME = 'index.json'  # TODO replace hardcoded `index.json` throughout the library with a reference to this variable.
    _indexSchema = {
        "$schema": "http://json-schema.org/schema#",
        '$id': 'extraReflectionIndexSchema',
        'title': 'extraReflectionIndexSchema',
        'type': 'object',
        'properties': {
            'reflectanceCubes': {
                'type': 'array',
                'items': {
                    'type': 'object',
                    'properties': {
                        'fileName': {'type': 'string'},
                        'description': {'type': 'string'},
                        'idTag': {'type': 'string'},
                        'name': {'type': 'string'},
                        'md5': {'type': 'string'}
                    },
                    'required': ['fileName', 'description', 'idTag', 'name']
                }
            },
            'creationDate': {'type': 'string'}
        }
    }

    def __init__(self, cubes: List[ERIndexCube], creationDate: Optional[str] = None):
        self.cubes = cubes
        if creationDate is None:
            self.creationDate = datetime.strftime(datetime.now(), dateTimeFormat)
        else:
            self.creationDate = creationDate

    @classmethod
    def load(cls, f: TextIO) -> ERIndex:
        try:
            indexFile = json.load(f)  # For unknown reasons this can sometime raises a JsonDecodeError and the index file needs to be redownloaded.
        except json.JSONDecodeError as e:
            f.seek(0)
            logging.getLogger(__name__).error(f"ERIndex.load() encountered a json.JSONDecodeError. File contents:\n {f.read()}")
            raise e  # Other parts of the application rely on this error to determine what to do.
        jsonschema.validate(indexFile, schema=cls._indexSchema)
        cubes = [ERIndexCube.fromDict(i) for i in indexFile['reflectanceCubes']]
        return cls(cubes, indexFile['creationDate'])

    def toDict(self) -> dict:
        return {'creationDate': self.creationDate, 'reflectanceCubes': [i.toDict() for i in self.cubes]}

    def toJson(self, directory: str):
        with open(os.path.join(directory, self.FILENAME), 'w') as f:
            json.dump(self.toDict(), f, indent=4)

    def getItemFromIdTag(self, idTag: str) -> ERIndexCube:
        for i in self.cubes:
            if i.idTag == idTag:
                return i
        raise ValueError(f"No item with idTag {idTag} was found.")

    @classmethod
    def merge(cls, index1: ERIndex, index2: ERIndex) -> ERIndex:
        """
        Provided two ERIndex objects this method will return a new ERIndex that merges the entries from both indices.
        If both indices contain an ERIndexCube entry that is similar but not identical a value error will be raised.
        Args:
            index1: The first ERIndex object
            index2: The seconds ERIndex object

        Returns:
            A new ERIndex object that combines the entries from both inputs.
        """
        import itertools
        newCubes = []
        idx1Cubes = [cube for cube in index1.cubes]
        idx2Cubes = [cube for cube in index2.cubes]
        for cube in idx1Cubes:
            match = [cube.idTag == c.idTag for c in idx2Cubes]
            if any(match): # The entry exists in both indexes, check for consistency
                matchingCube = next(itertools.compress(idx2Cubes, match))
                if cube != matchingCube:  # The idTags match but the entries are not identical
                    raise ValueError(f"The two indices contain entries that have the same idTags but do not match.")
                else:
                    newCubes.append(cube)
                    idx2Cubes.remove(matchingCube)
            else:  # index1 entry wasn't found in index2
                newCubes.append(cube)
        for cube in idx2Cubes:  # Everything still left in this list should be unique to index 2.
            newCubes.append(cube)
        return ERIndex(newCubes)


class ERIndexCube:
    """Information about a single ExtraReflectance data cube without necesarrily having the data.
    Useful for matching a local datafile/index entry with an online one."""
    def __init__(self, fileName: str, description: str, idTag: str, name: str, md5: Optional[str]):
        self.fileName = fileName
        self.description = description
        self.idTag = idTag
        self.name = name
        self.md5 = md5

    @classmethod
    def fromDict(cls, d: dict) -> ERIndexCube:
        return cls(**d)

    def toDict(self):
        return {'fileName': self.fileName, 'description': self.description, 'idTag': self.idTag, 'name': self.name, 'md5': self.md5}

    def __repr__(self) -> str:
        return f"ERIndexCube({self.toDict()})"

    def __eq__(self, other: ERIndex):
        assert isinstance(other, ERIndexCube)
        return self.toDict() == other.toDict()


# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
import os
from typing import Tuple, List, Optional
import typing
from PyQt5.QtCore import QThread

from pwspy_gui.PWSAnalysisApp._dockWidgets.AnalysisSettingsDock import AbstractRuntimeAnalysisSettings
from pwspy_gui.PWSAnalysisApp.sharedWidgets import ScrollableMessageBox
from pwspy_gui.sharedWidgets.dialogs import BusyDialog
from PyQt5 import QtCore
from PyQt5.QtWidgets import QMessageBox, QInputDialog
import pwspy.dataTypes as pwsdt
from pwspy.analysis import AbstractAnalysisSettings, AbstractAnalysis
from pwspy.analysis.dynamics import DynamicsAnalysis
from pwspy.analysis.pws import PWSAnalysis
from pwspy_gui.PWSAnalysisApp._dockWidgets.AnalysisSettingsDock.runtimeSettings import PWSRuntimeAnalysisSettings, DynamicsRuntimeAnalysisSettings
from pwspy.analysis.warnings import AnalysisWarning
from pwspy.dataTypes import ICRawBase, PwsMetaData, DynMetaData
from pwspy.utility.fileIO import loadAndProcess
if typing.TYPE_CHECKING:
    from pwspy_gui.PWSAnalysisApp.App import PWSApp


def safeCallback(func):
    """A decorator to make a function print its traceback without crashing."""
    def newFunc(*args):
        try:
            func(*args)
        except Exception as e:
            logger = logging.getLogger(__name__)
            logger.exception(e)
    return newFunc


class AnalysisManager(QtCore.QObject):
    analysisDone = QtCore.pyqtSignal(str, AbstractAnalysisSettings, list)

    def __init__(self, app: PWSApp):
        super().__init__()
        self.app = app

    def runList(self):
        """Run multiple queued analyses as specified by the user."""
        for anSettings in self.app.window.analysisSettings.getListedAnalyses():
            self.runSingle(anSettings)
            self.app.window.cellSelector.refreshCellItems()  # Refresh our displayed cell info

    @safeCallback
    def runSingle(self, anSettings: AbstractRuntimeAnalysisSettings) -> Tuple[str, AbstractAnalysisSettings, List[Tuple[List[AnalysisWarning], pwsdt.Acquisition]]]:
        """Run a single analysis batch"""
        logger = logging.getLogger(__name__)
        userSpecifiedBinning: Optional[int] = None
        cellMetas = anSettings.getCellMetadatas()
        refMeta = anSettings.getReferenceMetadata()
        cameraCorrection = anSettings.getSaveableSettings().cameraCorrection
        #Determine which cells already have an analysis by this name and raise a deletion dialog.
        conflictCells = []
        for cell in cellMetas:
            if anSettings.getAnalysisName() in cell.getAnalyses():
                conflictCells.append(cell)
        if len(conflictCells) > 0:
            ret = ScrollableMessageBox.question(self.app.window, "File Conflict", f"The following cells already have an analysis named {anSettings.getAnalysisName()}. Do you want to delete existing analyses and continue?: \n {', '.join([os.path.split(i.acquisitionDirectory.filePath)[-1] for i in conflictCells])}")
            if ret == QMessageBox.Yes:
                [cell.removeAnalysis(anSettings.getAnalysisName()) for cell in conflictCells]
            else:
                return
        if cameraCorrection is None:  # This means that the user has selected automatic cameraCorrection
            correctionsOk = self._checkAutoCorrectionConsistency(cellMetas + [refMeta])
        else:
            correctionsOk = True #We're using a user provided camera correction so we assume it's good to go.
        if correctionsOk:
            if isinstance(anSettings, PWSRuntimeAnalysisSettings):
                AnalysisClass = PWSAnalysis
                refMeta: PwsMetaData
            elif isinstance(anSettings, DynamicsRuntimeAnalysisSettings):
                AnalysisClass = DynamicsAnalysis
                refMeta: DynMetaData
            else:
                raise TypeError(f"Analysis settings of type: {type(anSettings)} are not supported.")
            ref = refMeta.toDataClass()
            if cameraCorrection is not None:
                try:
                    ref.correctCameraEffects(cameraCorrection)  # Apply the user-specified correction. This will fail if the image doesn't have binning metadata.
                except ValueError:
                    userSpecifiedBinning, pressedOk = QInputDialog.getInt(self.app.window, "Specify binning", "Please specify the camera binning that was used for these acquisitions.", 1, 1, 4)
                    if not pressedOk:  # User pressed cancel
                        return
                    ref.correctCameraEffects(cameraCorrection, binning=userSpecifiedBinning)
            else:
                logger.info("Using automatically detected camera corrections")
                ref.correctCameraEffects()
            if anSettings.getExtraReflectanceMetadata() is not None:  # if the ER is None, this means we are skipping the Extra reflection correction.
                if refMeta.systemName != anSettings.getExtraReflectanceMetadata().systemName:
                    ans = QMessageBox.question(self.app.window, "Uh Oh", f"The reference was acquired on system: {refMeta.systemName} while the extra reflectance correction was acquired on system: {anSettings.extraReflectanceMetadata.systemName}. Are you sure you want to continue?")
                    if ans == QMessageBox.No:
                        return
            logger.info("Initializing analysis")
            analysis = AnalysisClass(anSettings.getSaveableSettings(), anSettings.getExtraReflectanceMetadata(), ref)
            useParallelProcessing = self.app.parallelProcessing
            if (len(cellMetas) <= 3): #No reason to start 3 parallel processes for less than 3 cells.
                useParallelProcessing = False
            if useParallelProcessing:
                #Rather than copy arrays for each process, have read-only arrays that are shared between processes with shared memory. saves a few gigs of ram and speeds things up.
                logger.info("AnalysisManager: Using parallel processing. Creating shared memory.")
                try:
                    analysis.copySharedDataToSharedMemory()
                except NotImplementedError:
                    pass
            else:
                logger.info("Not using parallel processing.")
            #Run parallel/multithreaded processing
            t = self.AnalysisThread(cellMetas, analysis, anSettings.getAnalysisName(), cameraCorrection, userSpecifiedBinning, useParallelProcessing)
            b = BusyDialog(self.app.window, "Processing. Please Wait...")
            t.finished.connect(b.accept)

            def handleError(e: Exception, trace: str):
                logger.warning(trace)
                QMessageBox.information(self.app.window, "Oh No", str(e))
            t.errorOccurred.connect(handleError)
            t.start()
            b.exec()
            warnings = t.warnings
            warnings = [(warn, md) for warn, md in warnings if md is not None]
            ret = (anSettings.getAnalysisName(), anSettings.getSaveableSettings(), warnings)
            self.analysisDone.emit(*ret)
            return ret
        else:
            raise ValueError("Hmm. There appears to be a problem with different images using different `camera corrections`. Were all images taken on the same camera?")

    def _checkAutoCorrectionConsistency(self, cellMetas: List[pwsdt.AnalysisManagerMetaDataBase]) -> bool:
        """Confirm that all metadatas in cellMetas have identical camera corrections. otherwise we can't proceed"""
        camCorrections = [i.cameraCorrection for i in cellMetas]
        names = [os.path.split(i.filePath)[-1] for i in cellMetas]
        missing = []
        for name, cam in zip(names, camCorrections):
            if cam is None:
                missing.append(name)
        if len(missing) > 0:
            missingMessage = str(missing) if len(missing) <= 3 else 'Many cells are'
            QMessageBox.information(self.app.window, 'Hmm', f'{missingMessage} missing automatic camera correction')
            return False
        if len(set([hash(i) for i in camCorrections])) > 1:
            QMessageBox.information(self.app.window, 'Hmm', "Multiple camera corrections are present in the set of selected cells.")
            return False
        return True


    class AnalysisThread(QThread):
        errorOccurred = QtCore.pyqtSignal(Exception, str)

        def __init__(self, cellMetas, analysis, anName, cameraCorrection, userSpecifiedBinning, parallel):
            super().__init__()
            self.cellMetas = cellMetas
            self.analysis = analysis
            self.anName = anName
            self.cameraCorrection = cameraCorrection
            self.userSpecifiedBinning = userSpecifiedBinning
            self.warnings = None
            self.parallel = parallel

        def run(self):
            try:
                self.warnings = loadAndProcess(self.cellMetas, processorFunc=self._process, initArgs=[self.analysis, self.anName, self.cameraCorrection, self.userSpecifiedBinning],
                                     parallel=self.parallel, initializer=self._initializer) # Returns a list of Tuples, each tuple containing a list of warnings and the PwsMetaData to go with it.
            except Exception as e:
                import traceback
                trace = traceback.format_exc()
                self.errorOccurred.emit(e, trace)


        @staticmethod
        def _initializer(analysis: AbstractAnalysis, analysisName: str, cameraCorrection: pwsdt.CameraCorrection, userSpecifiedBinning: Optional[int] = None):
            """This method is run once for each process that is spawned. it initialized _resources that are shared between each iteration of _process."""
            global pwspyAnalysisAppParallelGlobals
            logger = logging.getLogger(__name__)
            logger.info('initializing!')
            pwspyAnalysisAppParallelGlobals = {'analysis': analysis, 'analysisName': analysisName,
                                               'cameraCorrection': cameraCorrection, 'binning': userSpecifiedBinning}

        @staticmethod
        def _process(im: pwsdt.ICRawBase):
            """This method is run in parallel. once for each acquisition data that we want to analyze.
            Returns a list of AnalysisWarnings objects with the associated metadat object"""
            global pwspyAnalysisAppParallelGlobals
            analysis = pwspyAnalysisAppParallelGlobals['analysis']
            analysisName = pwspyAnalysisAppParallelGlobals['analysisName']
            cameraCorrection = pwspyAnalysisAppParallelGlobals['cameraCorrection']
            userSpecifiedBinning = pwspyAnalysisAppParallelGlobals['binning']
            if cameraCorrection is not None:
                if userSpecifiedBinning is None:
                    im.correctCameraEffects(cameraCorrection)
                else:
                    im.correctCameraEffects(cameraCorrection, binning=userSpecifiedBinning)
            else:
                im.correctCameraEffects()
            results, warnings = analysis.run(im)
            if len(warnings) > 0:
                md = im.metadata
            else:
                md = None
            im.metadata.saveAnalysis(results, analysisName)
            return warnings, md




"""
This module is a plugin that adds functionality for selecting data that has been acquired using the new
"Acquisition Sequencer" in the PWS acquisition software.
"""

from ._plugin import AcquisitionSequencerPlugin





# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import typing as t_
import numpy as np
import pwspy.dataTypes as pwsdt
from PyQt5 import QtCore, QtGui
from PyQt5.QtCore import pyqtSignal, QObject
from PyQt5.QtWidgets import QWidget, QGridLayout, QButtonGroup, QPushButton, QDialog, QSpinBox, QLabel, \
    QMessageBox, QMenu, QAction, QApplication
from pwspy_gui.PWSAnalysisApp._roiManager import _DefaultROIManager
from pwspy_gui.PWSAnalysisApp.componentInterfaces import ROIManager

from pwspy_gui.PWSAnalysisApp.utilities.conglomeratedAnalysis import ConglomerateAnalysisResults
import os
from pwspy_gui.PWSAnalysisApp.sharedWidgets.plotting._analysisViewer import AnalysisViewer
from mpl_qt_viz.roiSelection import FullImPaintCreator, AdjustableSelector, LassoCreator, EllipseCreator, RegionalPaintCreator, PolygonModifier, WaterShedPaintCreator
if t_.TYPE_CHECKING:
    from pwspy.analysis.pws import PWSAnalysisResults
    from pwspy.analysis.dynamics import DynamicsAnalysisResults

    AnalysisResultsComboType = t_.Union[ConglomerateAnalysisResults,  # Internally we will use the ConglomerateAnalysisResults but we accept a regular tuple as well.
                                        t_.Tuple[
                                            t_.Optional[PWSAnalysisResults],
                                            t_.Optional[DynamicsAnalysisResults]
                                        ]]


class RoiDrawer(QWidget):
    """
    A widget for interactively drawing ROIs. Defaults to showing as it's own window, this can be overridden with the `flags` argument.

    Args:
        metadatas: A list of pwspy AcquisitionDirectory `Acquisition` objects paired with optional analysis results objects for that acquisition.
    """
    roiCreated = pyqtSignal(pwsdt.Acquisition, pwsdt.RoiFile, bool)  # Fired when a roi is created by this widget.
    roiDeleted = pyqtSignal(pwsdt.Acquisition, pwsdt.RoiFile)
    roiModified = pyqtSignal(pwsdt.Acquisition, pwsdt.RoiFile)
    metadataChanged = pyqtSignal(pwsdt.Acquisition)  # The acquisition we are looking at has been switched.

    def __init__(self, metadatas: t_.List[t_.Tuple[pwsdt.Acquisition, t_.Optional[AnalysisViewer.AnalysisResultsComboType]]],
                 roiManager: ROIManager = None, parent=None, flags=QtCore.Qt.Window,
                 title: str = "Roi Drawer 3000", initialField=AnalysisViewer.PlotFields.Thumbnail):
        QWidget.__init__(self, parent=parent, flags=flags)
        self.setWindowTitle(title)
        self.metadatas = metadatas

        layout = QGridLayout()

        self._mdIndex = 0
        self.roiManager = roiManager if roiManager else _DefaultROIManager(self)
        self.anViewer = AnalysisViewer(self.metadatas[self._mdIndex][0], self.metadatas[self._mdIndex][1], title, initialField=initialField, roiManager=self.roiManager)
        self.anViewer.roiPlot.roiDeleted.connect(lambda acq, roi: self.roiDeleted.emit(acq, roi))
        self.anViewer.roiPlot.roiModified.connect(lambda acq, roi: self.roiModified.emit(acq, roi))
        self.anViewer.roiPlot.roiCreated.connect(lambda acq, roi: self.roiCreated.emit(acq, roi, False))

        self.newRoiDlg = NewRoiDlg(self)

        self.noneButton = QPushButton("Inspect")
        self.lassoButton = QPushButton("Lasso")
        self.lassoButton.setToolTip(LassoCreator.getHelpText())
        self.ellipseButton = QPushButton("Ellipse")
        self.ellipseButton.setToolTip(EllipseCreator.getHelpText())
        self.paintButton = QPushButton("Paint")
        self.paintButton.setToolTip(RegionalPaintCreator.getHelpText())
        self.lastButton_ = None

        self.buttonGroup = QButtonGroup(self)
        self.buttonGroup.addButton(self.noneButton)
        self.buttonGroup.addButton(self.lassoButton)
        self.buttonGroup.addButton(self.ellipseButton)
        self.buttonGroup.addButton(self.paintButton)
        self.buttonGroup.buttonReleased.connect(self.handleButtons)
        [i.setCheckable(True) for i in self.buttonGroup.buttons()]
        self.noneButton.setChecked(True) #This doesn't seem to trigger handle buttons. we'll do that at the end of the constructor

        def handleAdjustButton(checkstate: bool):
            if self.selector is not None:
                self.selector.adjustable = checkstate

        self.adjustButton = QPushButton("Tune")
        self.adjustButton.setToolTip(PolygonModifier.getHelpText())
        self.adjustButton.setCheckable(True)
        self.adjustButton.setMaximumWidth(50)
        self.adjustButton.toggled.connect(handleAdjustButton)

        def showNextCell():
            idx = self._mdIndex + 1
            if idx >= len(self.metadatas):
                idx = 0
            self._updateDisplayedCell(idx)

        def showPreviousCell():
            idx = self._mdIndex - 1
            if idx < 0:
                idx = len(self.metadatas) - 1
            self._updateDisplayedCell(idx)

        self.previousButton = QPushButton('←')
        self.nextButton = QPushButton('→')
        self.previousButton.released.connect(showPreviousCell)
        self.nextButton.released.connect(showNextCell)

        layout.addWidget(self.noneButton, 0, 0, 1, 1)
        layout.addWidget(self.lassoButton, 0, 1, 1, 1)
        layout.addWidget(self.ellipseButton, 0, 2, 1, 1)
        layout.addWidget(self.paintButton, 0, 3, 1, 1)
        layout.addWidget(self.adjustButton, 0, 4, 1, 1)
        layout.addWidget(self.previousButton, 0, 6, 1, 1)
        layout.addWidget(self.nextButton, 0, 7, 1, 1)
        layout.addWidget(self.anViewer, 1, 0, 8, 8)
        self.setLayout(layout)
        self.selector: AdjustableSelector = AdjustableSelector(self.anViewer.roiPlot.ax, self.anViewer.roiPlot.im, LassoCreator, onfinished=self.finalizeRoi, onPolyTuningCancelled=lambda: self.selector.setActive(True))
        self.handleButtons(self.noneButton)  # Helps initialize state
        self.show()

    def finalizeRoi(self, verts: np.ndarray):
        roiName = self.anViewer.roiPlot.roiFilter.currentText()
        if roiName == '':
            QMessageBox.information(self, 'Wait', 'Please type an ROI name into the box at the top of the screen.')
            self.selector.setActive(True)
            return
        shape = self.anViewer.data.shape
        self.newRoiDlg.show()
        self.newRoiDlg.exec()
        if self.newRoiDlg.result() == QDialog.Accepted:
            md = self.metadatas[self._mdIndex][0]
            self._saveNewRoi(roiName, self.newRoiDlg.number, np.array(verts), shape, md)
        self.selector.setActive(True)  # Start the next roiFile.

    def _saveNewRoi(self, name: str, num: int, verts, datashape, acq: pwsdt.Acquisition):
        roi = pwsdt.Roi.fromVerts(verts, datashape)
        try:
            roiFile = self.roiManager.createRoi(acq, roi, name, num, overwrite=False)
            self.roiCreated.emit(acq, roiFile, False)
        except OSError as ose:
            ans = QMessageBox.question(self.anViewer, 'Overwrite?',
                                       f"Roi {name}:{num} already exists. Overwrite?")
            if ans == QMessageBox.Yes:
                roiFile = self.roiManager.getROI(acq, name, num)
                self.roiManager.updateRoi(roiFile, roi)
                self.roiCreated.emit(acq, roiFile, True)

    def handleButtons(self, button):
        if button is self.lassoButton and self.lastButton_ is not button:
            self.selector.setSelector(LassoCreator)
            self.selector.setActive(True)
            self.anViewer.roiPlot.enableHoverAnnotation(False)
            self.adjustButton.setEnabled(True)
        elif button is self.ellipseButton and self.lastButton_ is not button:
            self.selector.setSelector(EllipseCreator)
            self.selector.setActive(True)
            self.anViewer.roiPlot.enableHoverAnnotation(False)
            self.adjustButton.setEnabled(True)
        elif button is self.paintButton:
            def setSelector(sel):
                self.selector.setSelector(sel)
                self.selector.setActive(True)
                self.anViewer.roiPlot.enableHoverAnnotation(False)
                self.adjustButton.setEnabled(True)

            menu = QMenu(self)
            regionalAction = QAction("Regional")
            regionalAction.triggered.connect(lambda: setSelector(RegionalPaintCreator))
            menu.addAction(regionalAction)
            fullAction = QAction("Full Image")
            fullAction.triggered.connect(lambda: setSelector(FullImPaintCreator))
            menu.addAction(fullAction)
            watershedAction = QAction("Watershed")
            watershedAction.triggered.connect(lambda: setSelector(WaterShedPaintCreator))
            menu.addAction(watershedAction)
            menu.exec(self.mapToGlobal(self.paintButton.pos()))

        elif button is self.noneButton and self.lastButton_ is not button:
            if self.selector is not None:
                self.selector.setActive(False)
            self.anViewer.roiPlot.enableHoverAnnotation(True)
            self.adjustButton.setEnabled(False)
        self.lastButton_ = button

    def _updateDisplayedCell(self, idx: int):
        currRoi = self.anViewer.roiPlot.roiFilter.currentText()  # Since the next cell we look at will likely not have rois of the current name we want to manually force the ROI name to stay the same.
        md, analysis = self.metadatas[idx]
        self.anViewer.setMetadata(md, analysis=analysis)
        self.anViewer.roiPlot.roiFilter.setEditText(currRoi)  # manually force the ROI name to stay the same.
        self.selector.reset()  # Make sure to get rid of all rois
        self.setWindowTitle(f"Roi Drawer - {os.path.split(md.filePath)[-1]}")
        self.metadataChanged.emit(md)
        self._mdIndex = idx

    def setDisplayedAcquisition(self, acq: pwsdt.Acquisition):
        """Switch the image to display images associated with `acq`. If `acq` wasn't passed in to the constructor of this object then
        an IndexError will be raised.

        Args:
            acq: The acquisition to display.
        """
        for i, (mdAcq, analysis) in enumerate(self.metadatas):
            if mdAcq is acq:
                self._updateDisplayedCell(i)
                return
        raise IndexError(f"Acquisition {acq} was not found in the list of available images.")

    def closeEvent(self, a0: QtGui.QCloseEvent) -> None:
        self.selector.setActive(False)  # This cleans up remaining resources of the selector widgets.
        super().closeEvent(a0)


class NewRoiDlg(QDialog):
    def __init__(self, parent: RoiDrawer):
        super().__init__(parent=parent, flags=QtCore.Qt.FramelessWindowHint)
        self.setWindowTitle("ROI #")
        self.parent = parent
        self.setModal(True)
        self.number = None
        l = QGridLayout()
        self.numBox = QSpinBox()
        self.numBox.setMaximum(100000)
        self.numBox.setMinimum(0)
        self.okButton = QPushButton("Ok")
        self.okButton.released.connect(self.accept)
        self.cancelButton = QPushButton("Cancel")
        self.cancelButton.released.connect(self.reject)
        l.addWidget(QLabel("ROI #"), 0, 0, 1, 1, alignment=QtCore.Qt.AlignRight)
        l.addWidget(self.numBox, 0, 1, 1, 1)
        l.addWidget(self.okButton, 1, 0, 1, 1)
        l.addWidget(self.cancelButton, 1, 1, 1, 1)
        self.setLayout(l)

    def accept(self) -> None:
        self.number = self.numBox.value()
        self.numBox.setValue(self.numBox.value()+1)  # Increment the value.
        super().accept()

    def reject(self) -> None:
        self.number = None
        super().reject()

    def show(self) -> None:
        if len(self.parent.anViewer.roiPlot.rois) > 0:
            roiParams = self.parent.anViewer.roiPlot.rois
            newNum = max([param.roiFile.number for param in roiParams]) + 1  # Set the box 1 number abox the maximum found
            self.numBox.setValue(newNum)
        else:
            self.numBox.setValue(0)  # start at 0
        super().show()





from PyQt5.QtWidgets import QWidget
from pwspy_gui.PWSAnalysisApp.componentInterfaces import CellSelector
from pwspy_gui.PWSAnalysisApp.pluginInterfaces import CellSelectorPlugin
import typing as t_
import pwspy.dataTypes as pwsdt
from pwspy_gui.PWSAnalysisApp.plugins.backgroundAutoROI.drawBackgroundROIs import BGROIDialog, BGRoiDrawer


class BackgroundAutoROIPlugin(CellSelectorPlugin):
    def __init__(self):
        self._selector: CellSelector = None
        self._parentWidget = None

    def setContext(self, selector: CellSelector, parent: QWidget):
        """set the CellSelector that this plugin is associated to."""
        self._selector = selector
        self._parentWidget = parent

    def onCellsSelected(self, cells: t_.List[pwsdt.Acquisition]):
        """This method will be called when the CellSelector indicates that it has had new cells selected."""
        pass

    def onReferenceSelected(self, cell: pwsdt.Acquisition):
        """This method will be called when the CellSelector indicates that it has had a new reference selected."""
        pass

    def onNewCellsLoaded(self, cells: t_.List[pwsdt.Acquisition]):
        """This method will be called when the CellSelector indicates that new cells have been loaded to the selector."""
        pass

    def getName(self) -> str:
        """The name to refer to this plugin by."""
        return "Auto Background ROI"

    def onPluginSelected(self):
        """This method will be called when the plugin is activated."""
        drawer = BGRoiDrawer(parent=self._parentWidget, roiManager=self._selector.getRoiManager())
        drawer.run(self._selector.getSelectedCellMetas())

    def additionalColumnNames(self) -> t_.Sequence[str]:
        """The header names for each column."""
        return tuple()

    def getTableWidgets(self, acq: pwsdt.Acquisition) -> t_.Sequence[QWidget]:
        """provide a widget for each additional column to represent `acq`"""
        return tuple()



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from PyQt5.QtWidgets import QDialog, QWidget, QGridLayout, QLabel, QProgressBar, QApplication
from PyQt5 import QtCore


class BusyDialog(QDialog):
    def __init__(self, parent: QWidget, msg: str, progressBar: bool = False):
        super().__init__(parent, QtCore.Qt.FramelessWindowHint)
        self.setModal(True)
        layout = QGridLayout()
        label = QLabel(msg)
        font = label.font()
        font.setBold(True)
        font.setPointSize(18)
        label.setFont(font)
        layout.addWidget(label)
        self.progressBar = QProgressBar() if progressBar else None
        if self.progressBar:
            layout.addWidget(self.progressBar)
        self.setLayout(layout)
        self.show()

    def setProgress(self, percent: int):
        if self.progressBar:
            self.progressBar.setValue(percent)


if __name__ == '__main__':
    import sys
    app = QApplication(sys.argv)
    dlg = BusyDialog(None, "Busy, Please Wait")
    sys.exit(app.exec())
